from pytest_mock import MockerFixture

from mrok.agent import ziticorn


def test_run_ziticorn(mocker: MockerFixture):
    fake_app = mocker.MagicMock()

    mocked_config = mocker.MagicMock()
    mocked_config_ctor = mocker.patch(
        "mrok.agent.ziticorn.MrokBackendConfig",
        return_value=mocked_config,
    )

    mocked_server = mocker.MagicMock()
    mocked_server_ctor = mocker.patch(
        "mrok.agent.ziticorn.MrokServer",
        return_value=mocked_server,
    )

    ziticorn.run_ziticorn(fake_app, "ziti-identity.json")

    mocked_config_ctor.assert_called_once_with(fake_app, "ziti-identity.json")
    mocked_server_ctor.assert_called_once_with(mocked_config)
    mocked_server.run.assert_called_once()


def test_run(mocker: MockerFixture):
    fake_app = mocker.MagicMock()
    mocked_start_fn = mocker.MagicMock()
    mocked_partial = mocker.patch("mrok.agent.ziticorn.partial", return_value=mocked_start_fn)

    mocked_master = mocker.MagicMock()
    mocked_master_ctor = mocker.patch(
        "mrok.agent.ziticorn.Master",
        return_value=mocked_master,
    )

    ziticorn.run(fake_app, "ziti-identity.json", workers=10, reload=True)

    mocked_partial.assert_called_once_with(
        ziticorn.run_ziticorn,
        fake_app,
        "ziti-identity.json",
    )
    mocked_master_ctor.assert_called_once_with(mocked_start_fn, workers=10, reload=True)
    mocked_master.run.assert_called_once()



import asyncio
import re
from typing import Annotated

import typer
from rich import print

from mrok.cli.commands.admin.utils import parse_tags
from mrok.conf import Settings
from mrok.ziti.api import ZitiManagementAPI
from mrok.ziti.services import register_extension

RE_EXTENSION_ID = re.compile(r"(?i)EXT-\d{4}-\d{4}")


async def do_register(settings: Settings, extension_id: str, tags: list[str] | None):
    async with ZitiManagementAPI(settings) as api:
        await register_extension(settings, api, extension_id, tags=parse_tags(tags))


def validate_extension_id(extension_id: str) -> str:
    if not RE_EXTENSION_ID.fullmatch(extension_id):
        raise typer.BadParameter("ext_id must match EXT-xxxx-yyyy (case-insensitive)")
    return extension_id


def register(app: typer.Typer) -> None:
    @app.command("extension")
    def register_extension(
        ctx: typer.Context,
        extension_id: str = typer.Argument(
            ..., callback=validate_extension_id, help="Extension ID in format EXT-xxxx-yyyy"
        ),
        tags: Annotated[
            list[str] | None,
            typer.Option(
                "--tag",
                "-t",
                help="Add tag",
                show_default=True,
            ),
        ] = None,
    ):
        """Register a new Extension in OpenZiti (service)."""
        asyncio.run(do_register(ctx.obj, extension_id, tags))
        print(f"🍻 [green]Extension [bold]{extension_id}[/bold] registered.[/green]")



from urllib.parse import quote

import pytest
from dynaconf.base import LazySettings
from httpx import AsyncClient
from pytest_httpx import HTTPXMock
from pytest_mock import MockerFixture

from mrok.ziti.api import ZitiManagementAPI
from mrok.ziti.constants import MROK_VERSION_TAG_NAME
from mrok.ziti.errors import (
    ConfigTypeNotFoundError,
    MrokError,
    ProxyIdentityNotFoundError,
    ServiceAlreadyRegisteredError,
    ServiceNotFoundError,
)
from tests.conftest import SettingsFactory


@pytest.mark.asyncio
async def test_list_extensions(
    api_client: AsyncClient, settings_factory: SettingsFactory, httpx_mock: HTTPXMock
):
    settings = settings_factory()
    httpx_mock.add_response(
        method="GET",
        url=f"{settings.ziti.api.management}/edge/management/v1/services?limit=5&offset=0",
        json={
            "meta": {"pagination": {"totalCount": 10, "limit": 5, "offset": 0}},
            "data": [{"id": f"svc{i}", "name": "svc"} for i in range(5)],
        },
    )
    httpx_mock.add_response(
        method="GET",
        url=f"{settings.ziti.api.management}/edge/management/v1/services?limit=5&offset=5",
        json={
            "meta": {"pagination": {"totalCount": 10, "limit": 5, "offset": 5}},
            "data": [{"id": f"svc{5 + i}", "name": "svc"} for i in range(5)],
        },
    )
    response = await api_client.get("/extensions?limit=5&offset=0")
    assert response.status_code == 200
    page_1 = response.json()
    assert len(page_1["data"]) == 5
    assert page_1["$meta"]["pagination"]["offset"] == 0

    response = await api_client.get("/extensions?limit=5&offset=5")
    assert response.status_code == 200
    page_2 = response.json()
    assert len(page_2["data"]) == 5
    assert page_2["$meta"]["pagination"]["offset"] == 5

    assert sorted(page_1["data"], key=lambda x: x["id"]) != sorted(
        page_2["data"], key=lambda x: x["id"]
    )


@pytest.mark.asyncio
async def test_register_extension(mocker: MockerFixture, api_client: AsyncClient):
    mocked_register = mocker.patch(
        "mrok.controller.routes.register_extension",
        return_value={
            "id": "a1b2cd",
            "name": "ext-1234-5678",
            "tags": {
                MROK_VERSION_TAG_NAME: "0.0.0.dev0",
                "account": "ACC-1234-5678",
            },
        },
    )

    response = await api_client.post(
        "/extensions",
        json={
            "extension": {"id": "EXT-1234-5678"},
            "tags": {"account": "ACC-1234-5678"},
        },
    )
    assert response.status_code == 201
    assert response.json() == {
        "id": "a1b2cd",
        "extension": {"id": "EXT-1234-5678"},
        "name": "ext-1234-5678",
        "tags": {
            MROK_VERSION_TAG_NAME: "0.0.0.dev0",
            "account": "ACC-1234-5678",
        },
    }
    assert mocked_register.call_count == 1
    assert isinstance(mocked_register.mock_calls[0].args[0], LazySettings)
    assert isinstance(mocked_register.mock_calls[0].args[1], ZitiManagementAPI)
    assert mocked_register.mock_calls[0].args[2] == "EXT-1234-5678"
    assert mocked_register.mock_calls[0].args[3] == {"account": "ACC-1234-5678"}


@pytest.mark.asyncio
async def test_register_extension_already_exists(mocker: MockerFixture, api_client: AsyncClient):
    mocker.patch(
        "mrok.controller.routes.register_extension",
        side_effect=ServiceAlreadyRegisteredError("Extension `EXT-1234-5678` already registered."),
    )

    response = await api_client.post(
        "/extensions",
        json={
            "extension": {"id": "EXT-1234-5678"},
            "tags": {"account": "ACC-1234-5678"},
        },
    )
    assert response.status_code == 400
    assert response.json() == {"detail": "Extension `EXT-1234-5678` already registered."}


@pytest.mark.asyncio
@pytest.mark.parametrize("exc_type", [ProxyIdentityNotFoundError, ConfigTypeNotFoundError])
async def test_register_extension_mrok_not_configured(
    mocker: MockerFixture,
    api_client: AsyncClient,
    exc_type: MrokError,
):
    mocker.patch(
        "mrok.controller.routes.register_extension",
        side_effect=exc_type("this is the error."),  # type: ignore
    )

    response = await api_client.post(
        "/extensions",
        json={
            "extension": {"id": "EXT-1234-5678"},
            "tags": {"account": "ACC-1234-5678"},
        },
    )
    assert response.status_code == 400
    assert response.json() == {"detail": "OpenZiti not configured properly: this is the error."}


@pytest.mark.asyncio
async def test_get_extension(
    settings_factory: SettingsFactory,
    api_client: AsyncClient,
    httpx_mock: HTTPXMock,
):
    settings = settings_factory()
    query = quote(
        f'(id="EXT-1234-5678" or name="ext-1234-5678") and tags.{MROK_VERSION_TAG_NAME} != null'
    )
    httpx_mock.add_response(
        method="GET",
        url=f"{settings.ziti.api.management}/edge/management/v1/services?filter={query}",
        json={
            "meta": {"pagination": {"totalCount": 1}},
            "data": [
                {
                    "id": "svc1",
                    "name": "ext-1234-5678",
                    "tags": {MROK_VERSION_TAG_NAME: "0.0.0.dev0"},
                }
            ],
        },
    )

    response = await api_client.get("/extensions/EXT-1234-5678")
    assert response.status_code == 200
    assert response.json() == {
        "id": "svc1",
        "extension": {"id": "EXT-1234-5678"},
        "name": "ext-1234-5678",
        "tags": {
            MROK_VERSION_TAG_NAME: "0.0.0.dev0",
        },
    }


@pytest.mark.asyncio
async def test_get_extension_not_found(
    settings_factory: SettingsFactory,
    api_client: AsyncClient,
    httpx_mock: HTTPXMock,
):
    settings = settings_factory()
    query = quote(
        f'(id="EXT-1234-5678" or name="ext-1234-5678") and tags.{MROK_VERSION_TAG_NAME} != null'
    )
    httpx_mock.add_response(
        method="GET",
        url=f"{settings.ziti.api.management}/edge/management/v1/services?filter={query}",
        json={
            "meta": {"pagination": {"totalCount": 0}},
            "data": [],
        },
    )

    response = await api_client.get("/extensions/EXT-1234-5678")
    assert response.status_code == 404


@pytest.mark.asyncio
async def test_delete_extension(
    mocker: MockerFixture,
    api_client: AsyncClient,
):
    mocker.patch(
        "mrok.controller.routes.unregister_extension",
    )

    response = await api_client.delete("/extensions/EXT-1234-5678")
    assert response.status_code == 204


@pytest.mark.asyncio
async def test_delete_extension_not_found(
    mocker: MockerFixture,
    api_client: AsyncClient,
):
    mocker.patch(
        "mrok.controller.routes.unregister_extension", side_effect=ServiceNotFoundError("not found")
    )

    response = await api_client.delete("/extensions/EXT-1234-5678")
    assert response.status_code == 404



from mrok.controller.dependencies.conf import AppSettings
from mrok.controller.dependencies.ziti import ZitiClientAPI, ZitiManagementAPI

__all__ = ["AppSettings", "ZitiClientAPI", "ZitiManagementAPI"]



import multiprocessing

from pytest_mock import MockerFixture
from typer.testing import CliRunner

from mrok.cli import app
from mrok.controller.app import app as fastapi_app


def test_run(mocker: MockerFixture):
    mocker.patch(
        "mrok.cli.commands.controller.run.get_logging_config",
        return_value={"logging": "config"},
    )
    mocked_app = mocker.MagicMock()
    mocked_standalone_app = mocker.patch(
        "mrok.cli.commands.controller.run.StandaloneApplication", return_value=mocked_app
    )
    runner = CliRunner()
    result = runner.invoke(app, ["controller", "run"])
    assert result.exit_code == 0
    mocked_standalone_app.assert_called_once_with(
        fastapi_app,
        {
            "bind": "127.0.0.1:8000",
            "workers": (multiprocessing.cpu_count() * 2) + 1,
            "worker_class": "uvicorn_worker.UvicornWorker",
            "logconfig_dict": {"logging": "config"},
            "reload": False,
        },
    )
    mocked_app.run.assert_called_once()


def test_run_with_options(mocker: MockerFixture):
    mocker.patch(
        "mrok.cli.commands.controller.run.get_logging_config",
        return_value={"logging": "config"},
    )
    mocked_app = mocker.MagicMock()
    mocked_standalone_app = mocker.patch(
        "mrok.cli.commands.controller.run.StandaloneApplication", return_value=mocked_app
    )
    runner = CliRunner()
    result = runner.invoke(
        app,
        ["controller", "run", "--host", "0.0.0.0", "--port", "8080", "--workers", "2", "--reload"],
    )
    assert result.exit_code == 0
    mocked_standalone_app.assert_called_once_with(
        fastapi_app,
        {
            "bind": "0.0.0.0:8080",
            "workers": 2,
            "worker_class": "uvicorn_worker.UvicornWorker",
            "logconfig_dict": {"logging": "config"},
            "reload": True,
        },
    )



from pytest_mock import MockerFixture

from mrok.http.protocol import HttpToolsProtocol, MrokHttpToolsProtocol


def test_protocol(mocker: MockerFixture):
    mocked_super_init = mocker.patch.object(HttpToolsProtocol, "__init__")
    proto = MrokHttpToolsProtocol("10", test=True)
    mocked_super_init.assert_called_once_with("10", test=True)
    assert proto.access_logger.name == "mrok.access"
    assert proto.logger.name == "mrok.proxy"
    assert proto.access_log == proto.access_logger.hasHandlers()



import asyncio
import contextlib
from functools import partial
from pathlib import Path

from mrok.agent.sidecar.app import ForwardApp
from mrok.http.config import MrokBackendConfig
from mrok.http.master import Master
from mrok.http.server import MrokServer


def run_sidecar(identity_file: str, target_addr: str | Path | tuple[str, int]):
    config = MrokBackendConfig(ForwardApp(target_addr), identity_file)
    server = MrokServer(config)
    with contextlib.suppress(KeyboardInterrupt, asyncio.CancelledError):
        server.run()


def run(
    identity_file: str,
    target_addr: str | Path | tuple[str, int],
    workers=4,
    reload=False,
):
    start_fn = partial(run_sidecar, identity_file, target_addr)
    master = Master(start_fn, workers=workers, reload=reload)
    master.run()



from __future__ import annotations

from collections.abc import Sequence
from typing import Any

from fastapi import Query
from fastapi_pagination import create_page, resolve_params
from fastapi_pagination.bases import AbstractPage, AbstractParams, RawParams
from pydantic import BaseModel, Field

from mrok.controller.schemas import BaseSchema
from mrok.ziti.api import BaseZitiAPI


class MetaPagination(BaseModel):
    limit: int
    offset: int
    total: int


class Meta(BaseModel):
    pagination: MetaPagination


class LimitOffsetParams(BaseModel, AbstractParams):
    limit: int = Query(50, ge=0, le=1000, description="Page size limit")
    offset: int = Query(0, ge=0, description="Page offset")

    def to_raw_params(self) -> RawParams:
        return RawParams(  # pragma: no cover
            limit=self.limit,
            offset=self.offset,
        )


class LimitOffsetPage[S: BaseSchema](AbstractPage[S]):
    data: list[S]
    meta: Meta = Field(alias="$meta")

    __params_type__ = LimitOffsetParams  # type: ignore

    @classmethod
    def create(
        cls,
        items: Sequence[S],
        params: AbstractParams,
        *,
        total: int | None = None,
        **kwargs: Any,
    ) -> LimitOffsetPage[S]:
        if not isinstance(params, LimitOffsetParams):  # pragma: no cover
            raise TypeError("params must be of type LimitOffsetParams")
        return cls(  # type: ignore
            data=items,
            meta=Meta(
                pagination=MetaPagination(
                    limit=params.limit,
                    offset=params.offset,
                    total=total,
                )
            ),
        )


async def paginate[S: BaseSchema](
    api: BaseZitiAPI,
    endpoint: str,
    schema_cls: type[S],
    extra_params: dict | None = None,
) -> AbstractPage[S]:
    params: LimitOffsetParams = resolve_params()
    page = await api.get_page(endpoint, params.limit, params.offset, extra_params)
    pagination_meta = page["meta"]["pagination"]
    total = pagination_meta["totalCount"]
    return create_page(
        [schema_cls(**item) for item in page["data"]],
        params=params,
        total=total,
    )



import pytest
from pytest_mock import MockerFixture

from mrok.ziti.constants import MROK_SERVICE_TAG_NAME, MROK_VERSION_TAG_NAME
from mrok.ziti.errors import (
    ProxyIdentityAlreadyExistsError,
    ServiceNotFoundError,
    UserIdentityNotFoundError,
)
from mrok.ziti.identities import enroll_proxy_identity, register_instance, unregister_instance


@pytest.mark.asyncio
async def test_register_instance(mocker: MockerFixture):
    mocked_jwt_decode = mocker.patch(
        "mrok.ziti.identities.jwt.decode", return_value={"jti": "jti-value"}
    )
    mocked_get_ca_certs = mocker.patch(
        "mrok.ziti.identities.pki.get_ca_certificates",
        return_value="ca-certificates-chain",
    )
    mocked_mgmt_api = mocker.AsyncMock()
    mocked_mgmt_api.search_service.return_value = {"id": "svc1"}
    mocked_mgmt_api.search_identity.return_value = None
    mocked_mgmt_api.create_user_identity.return_value = "identity-id"
    mocked_mgmt_api.get_identity.return_value = {
        "id": "identity_id",
        "name": "identity_name",
        "tags": {
            MROK_VERSION_TAG_NAME: "0.0.0.dev0",
            MROK_SERVICE_TAG_NAME: "svc1",
            "account": "ACC-1234",
        },
        "enrollment": {"ott": {"jwt": "enroll-jwt-token"}},
    }

    mocked_client_api = mocker.AsyncMock()
    mocked_client_api.base_url = "https://ziti.api"
    mocked_client_api.enroll_identity.return_value = {"data": {"cert": "identity-certificate"}}

    identity, identity_json = await register_instance(
        mocked_mgmt_api,
        mocked_client_api,
        "EXT-1234-5678",
        "INS-1234-5678-0001",
        tags={"account": "ACC-1234"},
    )
    assert identity == {
        "id": "identity_id",
        "name": "identity_name",
        "tags": {
            MROK_VERSION_TAG_NAME: "0.0.0.dev0",
            MROK_SERVICE_TAG_NAME: "svc1",
            "account": "ACC-1234",
        },
        "enrollment": {"ott": {"jwt": "enroll-jwt-token"}},
    }

    mocked_mgmt_api.search_service.assert_awaited_once_with("ext-1234-5678")
    mocked_mgmt_api.search_identity.assert_awaited_once_with("ins-1234-5678-0001.ext-1234-5678")
    mocked_mgmt_api.create_user_identity.assert_awaited_once_with(
        "ins-1234-5678-0001.ext-1234-5678",
        tags={
            MROK_SERVICE_TAG_NAME: "ext-1234-5678",
            "account": "ACC-1234",
        },
    )
    mocked_mgmt_api.get_identity.assert_awaited_once_with("identity-id")
    mocked_jwt_decode.assert_called_once_with(
        "enroll-jwt-token",
        algorithms=["RS256"],
        options={"verify_signature": False},
    )
    assert mocked_client_api.enroll_identity.mock_calls[0].args[0] == "jti-value"
    assert (
        mocked_client_api.enroll_identity.mock_calls[0]
        .args[1]
        .startswith("-----BEGIN CERTIFICATE REQUEST-----")
    )
    mocked_get_ca_certs.assert_awaited_once_with(mocked_mgmt_api)
    assert identity_json["ztAPI"] == "https://ziti.api"
    assert identity_json["id"]["key"].startswith("pem:-----BEGIN PRIVATE KEY-----")
    assert identity_json["id"]["cert"] == "pem:identity-certificate"
    assert identity_json["id"]["ca"] == "pem:ca-certificates-chain"

    mocked_mgmt_api.create_bind_service_policy.assert_awaited_once_with(
        "ins-1234-5678-0001.ext-1234-5678:bind",
        "svc1",
        "identity-id",
    )
    mocked_mgmt_api.create_router_policy.assert_awaited_once_with(
        "ins-1234-5678-0001.ext-1234-5678",
        "identity-id",
    )


@pytest.mark.asyncio
async def test_register_instance_service_not_found(mocker: MockerFixture):
    mocked_mgmt_api = mocker.AsyncMock()
    mocked_mgmt_api.search_service.return_value = None

    mocked_client_api = mocker.AsyncMock()

    with pytest.raises(ServiceNotFoundError) as cv:
        await register_instance(
            mocked_mgmt_api,
            mocked_client_api,
            "EXT-1234-5678",
            "INS-1234-5678-0001",
            tags={"account": "ACC-1234"},
        )

    assert str(cv.value) == "A service with name `EXT-1234-5678` does not exists."


@pytest.mark.asyncio
async def test_register_instance_identity_exists(mocker: MockerFixture):
    mocked_jwt_decode = mocker.patch(
        "mrok.ziti.identities.jwt.decode", return_value={"jti": "jti-value"}
    )
    mocked_get_ca_certs = mocker.patch(
        "mrok.ziti.identities.pki.get_ca_certificates",
        return_value="ca-certificates-chain",
    )
    mocked_mgmt_api = mocker.AsyncMock()
    mocked_mgmt_api.search_service.return_value = {"id": "svc1"}
    mocked_mgmt_api.search_identity.return_value = {"id": "identity-id"}
    mocked_mgmt_api.search_service_policy.return_value = {"id": "service-policy-id"}
    mocked_mgmt_api.search_router_policy.return_value = {"id": "router-policy-id"}

    mocked_mgmt_api.create_user_identity.return_value = "identity-id"
    mocked_mgmt_api.get_identity.return_value = {"enrollment": {"ott": {"jwt": "enroll-jwt-token"}}}

    mocked_client_api = mocker.AsyncMock()
    mocked_client_api.base_url = "https://ziti.api"
    mocked_client_api.enroll_identity.return_value = {"data": {"cert": "identity-certificate"}}

    _, identity_json = await register_instance(
        mocked_mgmt_api,
        mocked_client_api,
        "EXT-1234-5678",
        "INS-1234-5678-0001",
        tags={"account": "ACC-1234"},
    )

    mocked_mgmt_api.search_service.assert_awaited_once_with("ext-1234-5678")
    mocked_mgmt_api.search_identity.assert_awaited_once_with("ins-1234-5678-0001.ext-1234-5678")
    mocked_mgmt_api.create_user_identity.assert_awaited_once_with(
        "ins-1234-5678-0001.ext-1234-5678",
        tags={
            MROK_SERVICE_TAG_NAME: "ext-1234-5678",
            "account": "ACC-1234",
        },
    )
    mocked_mgmt_api.get_identity.assert_awaited_once_with("identity-id")
    mocked_jwt_decode.assert_called_once_with(
        "enroll-jwt-token",
        algorithms=["RS256"],
        options={"verify_signature": False},
    )
    assert mocked_client_api.enroll_identity.mock_calls[0].args[0] == "jti-value"
    assert (
        mocked_client_api.enroll_identity.mock_calls[0]
        .args[1]
        .startswith("-----BEGIN CERTIFICATE REQUEST-----")
    )
    mocked_get_ca_certs.assert_awaited_once_with(mocked_mgmt_api)
    assert identity_json["ztAPI"] == "https://ziti.api"
    assert identity_json["id"]["key"].startswith("pem:-----BEGIN PRIVATE KEY-----")
    assert identity_json["id"]["cert"] == "pem:identity-certificate"
    assert identity_json["id"]["ca"] == "pem:ca-certificates-chain"

    mocked_mgmt_api.create_bind_service_policy.assert_awaited_once_with(
        "ins-1234-5678-0001.ext-1234-5678:bind",
        "svc1",
        "identity-id",
    )
    mocked_mgmt_api.create_router_policy.assert_awaited_once_with(
        "ins-1234-5678-0001.ext-1234-5678",
        "identity-id",
    )
    mocked_mgmt_api.search_service_policy.assert_awaited_once_with(
        "ins-1234-5678-0001.ext-1234-5678:bind"
    )
    mocked_mgmt_api.delete_service_policy.assert_awaited_once_with("service-policy-id")
    mocked_mgmt_api.search_router_policy.assert_awaited_once_with(
        "ins-1234-5678-0001.ext-1234-5678"
    )
    mocked_mgmt_api.delete_router_policy.assert_awaited_once_with("router-policy-id")
    mocked_mgmt_api.delete_identity.assert_awaited_once_with("identity-id")


@pytest.mark.asyncio
async def test_register_instance_identity_exists_service_router_doesnt(mocker: MockerFixture):
    mocked_jwt_decode = mocker.patch(
        "mrok.ziti.identities.jwt.decode", return_value={"jti": "jti-value"}
    )
    mocked_get_ca_certs = mocker.patch(
        "mrok.ziti.identities.pki.get_ca_certificates",
        return_value="ca-certificates-chain",
    )
    mocked_mgmt_api = mocker.AsyncMock()
    mocked_mgmt_api.search_service.return_value = {"id": "svc1"}
    mocked_mgmt_api.search_identity.return_value = {"id": "identity-id"}
    mocked_mgmt_api.search_service_policy.return_value = None
    mocked_mgmt_api.search_router_policy.return_value = None

    mocked_mgmt_api.create_user_identity.return_value = "identity-id"
    mocked_mgmt_api.get_identity.return_value = {"enrollment": {"ott": {"jwt": "enroll-jwt-token"}}}

    mocked_client_api = mocker.AsyncMock()
    mocked_client_api.base_url = "https://ziti.api"
    mocked_client_api.enroll_identity.return_value = {"data": {"cert": "identity-certificate"}}

    _, identity_json = await register_instance(
        mocked_mgmt_api,
        mocked_client_api,
        "EXT-1234-5678",
        "INS-1234-5678-0001",
        tags={"account": "ACC-1234"},
    )

    mocked_mgmt_api.search_service.assert_awaited_once_with("ext-1234-5678")
    mocked_mgmt_api.search_identity.assert_awaited_once_with("ins-1234-5678-0001.ext-1234-5678")
    mocked_mgmt_api.create_user_identity.assert_awaited_once_with(
        "ins-1234-5678-0001.ext-1234-5678",
        tags={
            MROK_SERVICE_TAG_NAME: "ext-1234-5678",
            "account": "ACC-1234",
        },
    )
    mocked_mgmt_api.get_identity.assert_awaited_once_with("identity-id")
    mocked_jwt_decode.assert_called_once_with(
        "enroll-jwt-token",
        algorithms=["RS256"],
        options={"verify_signature": False},
    )
    assert mocked_client_api.enroll_identity.mock_calls[0].args[0] == "jti-value"
    assert (
        mocked_client_api.enroll_identity.mock_calls[0]
        .args[1]
        .startswith("-----BEGIN CERTIFICATE REQUEST-----")
    )
    mocked_get_ca_certs.assert_awaited_once_with(mocked_mgmt_api)
    assert identity_json["ztAPI"] == "https://ziti.api"
    assert identity_json["id"]["key"].startswith("pem:-----BEGIN PRIVATE KEY-----")
    assert identity_json["id"]["cert"] == "pem:identity-certificate"
    assert identity_json["id"]["ca"] == "pem:ca-certificates-chain"

    mocked_mgmt_api.create_bind_service_policy.assert_awaited_once_with(
        "ins-1234-5678-0001.ext-1234-5678:bind",
        "svc1",
        "identity-id",
    )
    mocked_mgmt_api.create_router_policy.assert_awaited_once_with(
        "ins-1234-5678-0001.ext-1234-5678",
        "identity-id",
    )
    mocked_mgmt_api.search_service_policy.assert_awaited_once_with(
        "ins-1234-5678-0001.ext-1234-5678:bind"
    )
    mocked_mgmt_api.delete_service_policy.assert_not_awaited()
    mocked_mgmt_api.search_router_policy.assert_awaited_once_with(
        "ins-1234-5678-0001.ext-1234-5678"
    )
    mocked_mgmt_api.delete_router_policy.assert_not_awaited()
    mocked_mgmt_api.delete_identity.assert_awaited_once_with("identity-id")


@pytest.mark.asyncio
async def test_unregister_instance(
    mocker: MockerFixture,
):
    mocked_mgmt_api = mocker.AsyncMock()
    mocked_mgmt_api.search_service.return_value = {"id": "svc1"}
    mocked_mgmt_api.search_identity.return_value = {"id": "identity-id"}
    mocked_mgmt_api.search_service_policy.return_value = {"id": "service-policy-id"}
    mocked_mgmt_api.search_router_policy.return_value = {"id": "router-policy-id"}

    await unregister_instance(
        mocked_mgmt_api,
        "EXT-1234-5678",
        "INS-1234-5678-0001",
    )

    mocked_mgmt_api.search_service.assert_awaited_once_with("ext-1234-5678")
    mocked_mgmt_api.search_identity.assert_awaited_once_with("ins-1234-5678-0001.ext-1234-5678")

    mocked_mgmt_api.search_service_policy.assert_awaited_once_with(
        "ins-1234-5678-0001.ext-1234-5678:bind"
    )
    mocked_mgmt_api.delete_service_policy.assert_awaited_once_with("service-policy-id")
    mocked_mgmt_api.search_router_policy.assert_awaited_once_with(
        "ins-1234-5678-0001.ext-1234-5678"
    )
    mocked_mgmt_api.delete_router_policy.assert_awaited_once_with("router-policy-id")
    mocked_mgmt_api.delete_identity.assert_awaited_once_with("identity-id")


@pytest.mark.asyncio
async def test_unregister_instance_service_not_found(mocker: MockerFixture):
    mocked_mgmt_api = mocker.AsyncMock()
    mocked_mgmt_api.search_service.return_value = None

    with pytest.raises(ServiceNotFoundError) as cv:
        await unregister_instance(
            mocked_mgmt_api,
            "EXT-1234-5678",
            "INS-1234-5678-0001",
        )

    assert str(cv.value) == "A service with name `EXT-1234-5678` does not exists."


@pytest.mark.asyncio
async def test_unregister_instance_instance_not_found(mocker: MockerFixture):
    mocked_mgmt_api = mocker.AsyncMock()
    mocked_mgmt_api.search_service.return_value = {"id": "svc1"}
    mocked_mgmt_api.search_identity.return_value = None

    with pytest.raises(UserIdentityNotFoundError) as cv:
        await unregister_instance(
            mocked_mgmt_api,
            "EXT-1234-5678",
            "INS-1234-5678-0001",
        )

    assert str(cv.value) == "Instance `INS-1234-5678-0001` not found."


@pytest.mark.asyncio
async def test_unregister_instance_policies_not_found(
    mocker: MockerFixture,
):
    mocked_mgmt_api = mocker.AsyncMock()
    mocked_mgmt_api.search_service.return_value = {"id": "svc1"}
    mocked_mgmt_api.search_identity.return_value = {"id": "identity-id"}
    mocked_mgmt_api.search_service_policy.return_value = None
    mocked_mgmt_api.search_router_policy.return_value = None

    await unregister_instance(
        mocked_mgmt_api,
        "EXT-1234-5678",
        "INS-1234-5678-0001",
    )

    mocked_mgmt_api.search_service.assert_awaited_once_with("ext-1234-5678")
    mocked_mgmt_api.search_identity.assert_awaited_once_with("ins-1234-5678-0001.ext-1234-5678")

    mocked_mgmt_api.search_service_policy.assert_awaited_once_with(
        "ins-1234-5678-0001.ext-1234-5678:bind"
    )
    mocked_mgmt_api.delete_service_policy.assert_not_awaited()
    mocked_mgmt_api.search_router_policy.assert_awaited_once_with(
        "ins-1234-5678-0001.ext-1234-5678"
    )
    mocked_mgmt_api.delete_router_policy.assert_not_awaited()
    mocked_mgmt_api.delete_identity.assert_awaited_once_with("identity-id")


@pytest.mark.asyncio
async def test_enroll_proxy_identity(mocker: MockerFixture):
    mocked_jwt_decode = mocker.patch(
        "mrok.ziti.identities.jwt.decode", return_value={"jti": "jti-value"}
    )
    mocked_get_ca_certs = mocker.patch(
        "mrok.ziti.identities.pki.get_ca_certificates",
        return_value="ca-certificates-chain",
    )
    mocked_mgmt_api = mocker.AsyncMock()
    mocked_mgmt_api.search_identity.return_value = None
    mocked_mgmt_api.create_device_identity.return_value = "identity-id"
    mocked_mgmt_api.get_identity.return_value = {"enrollment": {"ott": {"jwt": "enroll-jwt-token"}}}

    mocked_client_api = mocker.AsyncMock()
    mocked_client_api.base_url = "https://ziti.api"
    mocked_client_api.enroll_identity.return_value = {"data": {"cert": "identity-certificate"}}

    _, identity_json = await enroll_proxy_identity(
        mocked_mgmt_api,
        mocked_client_api,
        "mrok-proxy",
        tags={"test": "tag"},
    )

    mocked_mgmt_api.search_identity.assert_awaited_once_with("mrok-proxy")
    mocked_mgmt_api.create_device_identity.assert_awaited_once_with(
        "mrok-proxy", tags={"test": "tag"}
    )
    mocked_mgmt_api.get_identity.assert_awaited_once_with("identity-id")
    mocked_jwt_decode.assert_called_once_with(
        "enroll-jwt-token",
        algorithms=["RS256"],
        options={"verify_signature": False},
    )
    assert mocked_client_api.enroll_identity.mock_calls[0].args[0] == "jti-value"
    assert (
        mocked_client_api.enroll_identity.mock_calls[0]
        .args[1]
        .startswith("-----BEGIN CERTIFICATE REQUEST-----")
    )
    mocked_get_ca_certs.assert_awaited_once_with(mocked_mgmt_api)
    assert identity_json["ztAPI"] == "https://ziti.api"
    assert identity_json["id"]["key"].startswith("pem:-----BEGIN PRIVATE KEY-----")
    assert identity_json["id"]["cert"] == "pem:identity-certificate"
    assert identity_json["id"]["ca"] == "pem:ca-certificates-chain"


@pytest.mark.asyncio
async def test_enroll_proxy_identity_already_exists(mocker: MockerFixture):
    mocked_mgmt_api = mocker.AsyncMock()
    mocked_mgmt_api.search_identity.return_value = {"id": "identity-id"}

    with pytest.raises(ProxyIdentityAlreadyExistsError) as cv:
        await enroll_proxy_identity(
            mocked_mgmt_api,
            mocker.AsyncMock(),
            "mrok-proxy",
            tags={"test": "tag"},
        )

    assert str(cv.value) == "A proxy identity with name `mrok-proxy` already exists."



import multiprocessing
from pathlib import Path
from typing import Annotated

import typer

# from app.logging import get_logging_config
from mrok.agent import ziticorn


def number_of_workers():
    return (multiprocessing.cpu_count() * 2) + 1


default_workers = number_of_workers()


def register(app: typer.Typer) -> None:
    @app.command("asgi")
    def run_asgi(
        app: str = typer.Argument(
            ...,
            help="ASGI application",
        ),
        identity_file: Path = typer.Argument(
            ...,
            help="Identity json file",
        ),
        workers: Annotated[
            int,
            typer.Option(
                "--workers",
                "-w",
                help=f"Number of workers. Default: {default_workers}",
                show_default=True,
            ),
        ] = default_workers,
        reload: Annotated[
            bool,
            typer.Option(
                "--reload",
                "-r",
                help="Enable auto-reload. Default: False",
                show_default=True,
            ),
        ] = False,
    ):
        """Run an ASGI application exposing it through OpenZiti network."""
        ziticorn.run(app, str(identity_file), workers=workers, reload=reload)



import logging
import shlex
import tempfile

import pytest
from pytest_mock import MockerFixture
from typer.testing import CliRunner

from mrok.cli import app
from mrok.cli.commands.admin.bootstrap import bootstrap
from tests.conftest import SettingsFactory


def test_bootstrap_command(
    mocker: MockerFixture,
    settings_factory: SettingsFactory,
):
    logger = logging.getLogger("mrok")
    logger.handlers.clear()
    logger.propagate = True

    settings = settings_factory()
    mocker.patch("mrok.cli.main.get_settings", return_value=settings)
    mock_bootstrap_coro = mocker.MagicMock()
    mock_bootstrap = mocker.MagicMock(return_value=mock_bootstrap_coro)

    mocker.patch("mrok.cli.commands.admin.bootstrap.bootstrap", mock_bootstrap)
    mock_run = mocker.patch(
        "mrok.cli.commands.admin.bootstrap.asyncio.run",
        return_value=({"id": "identity_id"}, {"identity": "json"}),
    )
    runner = CliRunner()

    # Run the command
    with tempfile.NamedTemporaryFile() as f:
        result = runner.invoke(
            app,
            shlex.split(f"admin bootstrap {f.name} --tag tag=tagvalue"),
        )
        f.seek(0)
        json_data = f.read()
        assert json_data == b'{"identity": "json"}'
        assert result.exit_code == 0

    mock_run.assert_called_once_with(mock_bootstrap_coro)
    mock_bootstrap.assert_called_once_with(
        settings,
        False,
        {"tag": "tagvalue"},
    )


@pytest.mark.asyncio
async def test_bootstrap(mocker: MockerFixture, settings_factory: SettingsFactory):
    settings = settings_factory()
    mocked_api = mocker.AsyncMock()
    mocked_api.__aenter__.return_value = mocked_api
    mocked_api_ctor = mocker.patch(
        "mrok.cli.commands.admin.bootstrap.ZitiManagementAPI",
        return_value=mocked_api,
    )
    mocked_cl_api = mocker.AsyncMock()
    mocked_cl_api.__aenter__.return_value = mocked_cl_api
    mocked_cl_api_ctor = mocker.patch(
        "mrok.cli.commands.admin.bootstrap.ZitiClientAPI",
        return_value=mocked_cl_api,
    )
    mocked_bootstrap_identity = mocker.patch("mrok.cli.commands.admin.bootstrap.bootstrap_identity")

    await bootstrap(settings, False, None)
    mocked_bootstrap_identity.assert_awaited_once_with(
        mocked_api,
        mocked_cl_api,
        settings.proxy.identity,
        settings.proxy.mode,
        False,
        None,
    )
    mocked_api_ctor.assert_called_once_with(settings)
    mocked_cl_api_ctor.assert_called_once_with(settings)



import json
from enum import Enum
from pathlib import Path
from typing import Annotated

import typer
import yaml

from mrok.controller.openapi import generate_openapi_spec


class OutputFormat(str, Enum):
    json = "json"
    yaml = "yaml"


def register(app: typer.Typer) -> None:
    @app.command("openapi")
    def generate_spec(
        ctx: typer.Context,
        output: Annotated[
            Path | None,
            typer.Option(
                "--output",
                "-o",
                help="Output file",
            ),
        ] = Path("mrok_openapi_spec.yml"),
        output_format: Annotated[
            OutputFormat,
            typer.Option(
                "--output-format",
                "-f",
                help="Output file format",
            ),
        ] = OutputFormat.yaml,
    ):
        """
        Generates the mrok controller OpenAPI spec file.
        """
        from mrok.controller.app import app

        dump_fn = json.dump if output_format == OutputFormat.json else yaml.dump
        spec = generate_openapi_spec(app, ctx.obj)

        with open(output, "w") as f:  # type: ignore
            dump_fn(spec, f, indent=2)



import json
import tempfile
from collections.abc import AsyncGenerator, Callable, Generator
from datetime import UTC, datetime, timedelta
from typing import Any

import jwt
import pytest
from asgi_lifespan import LifespanManager
from dynaconf import Dynaconf
from fastapi import FastAPI
from httpx import ASGITransport, AsyncClient
from pytest_httpx import HTTPXMock

from mrok.conf import Settings, get_settings

SettingsFactory = Callable[..., Settings]


@pytest.fixture(scope="session")
def settings_factory() -> SettingsFactory:
    def _get_settings(
        ziti: dict | None = None,
        logging: dict | None = None,
        pagination: dict | None = None,
        proxy: dict | None = None,
        auth: dict | None = None,
    ) -> Settings:
        ziti = ziti or {
            "api": {
                "management": "https://ziti.example.com",
                "client": "https://ziti.example.com",
            },
            "read_timeout": 10,
            "ssl_verify": True,
            "auth": {"username": "user", "password": "pass", "identity": None},
        }
        logging = logging or {
            "debug": True,
            "rich": False,
        }
        pagination = pagination or {"limit": 5}
        auth = auth or {
            "openid_config_url": "http://example.com/openid-configuration",
            "audience": "mrok-audience",
            "read_timeout": 10,
        }
        proxy = proxy or {
            "identity": "public",
            "mode": "zrok",
            "domain": "exts.s1.today",
        }
        settings = Dynaconf(
            environments=True,
            settings_files=[],
            ENV_FOR_DYNACONF="testing",
            ZITI=ziti,
            LOGGING=logging,
            PAGINATION=pagination,
            PROXY=proxy,
            AUTH=auth,
        )

        return settings

    return _get_settings


@pytest.fixture()
def ziti_bad_request_error() -> dict[str, Any]:
    return {
        "error": {
            "code": "error_code",
            "message": "error_message",
            "cause": {
                "field": "field_name",
                "reason": "field_error",
            },
        }
    }


@pytest.fixture()
def ziti_identity_json() -> dict[str, Any]:
    return {
        "ztAPI": "https://ziti.exts.platform.softwareone.com/edge/client/v1",
        "ztAPIs": None,
        "configTypes": None,
        "id": {
            "key": "pem:-----BEGIN PRIVATE KEY-----\n...\n-----END PRIVATE KEY-----\n",
            "cert": "pem:-----BEGIN CERTIFICATE-----\n...\n-----END CERTIFICATE-----\n",
            "ca": "pem:-----BEGIN CERTIFICATE-----\n...\n-----END CERTIFICATE-----\n",
        },
        "enableHa": None,
        "mrok": {"identity": "ins-1234-5678-0001.ext-1234-5678"},
    }


@pytest.fixture()
def ziti_identity_file(ziti_identity_json: dict[str, Any]) -> Generator[str, None, None]:
    with tempfile.NamedTemporaryFile("w", suffix="json") as f:
        json.dump(ziti_identity_json, f)
        f.seek(0)
        yield f.name


@pytest.fixture()
def jwt_signing_key() -> str:
    return """-----BEGIN PRIVATE KEY-----
MIIEvQIBADANBgkqhkiG9w0BAQEFAASCBKcwggSjAgEAAoIBAQDh0IIuMVuxXSq1
ndmO13RLpqv+3Ndyh0aEP8Lz3/zQeyliYCfW8HMoP8vjC9ftGKdEHLG4eoUs6NF+
8dYh3MRtDikl0oH/J+/6DFS3zqG+NNaMUpTgmKsrn1w5MN/NPQxPp/Djc4nAcAoi
V/c0w5fft5CLoLkLt5iL8y2mTuR0VRhZ6fS+hQiQMcSBajtr8jPLhem9spJ53N5x
X/OQNTkP712s9qSwCLd8SoFzNFYam2kHKs+nfMbXmk0PXbLjTb9Sk6pOa4Ey5OZU
//8oFyMnEltuRUH158WskpxSsxwZosPOsrFXcEdq5DHPQppc5H6IjwevVr8VZCZ9
y0DhqbpHAgMBAAECggEAa23BiQ1VVIeA2p9qkbDBtH3qJJlW7DccHq2g51nf0iVN
1m3tdi21c2gKbQ1E4BjS9q36BMxREEEA7bZKy5hWYJWUkNhZnRSYb+qu4TUuMKm9
ikt9ZW5sPJgXjWPJjUAmW70HdoYQelszDwyvYExPNBEF9M91SXRCYA5EYPL2b7rp
FjUzbq4/a4C9GYUP7EhBYYyZ+KnFDa9mxdbUFrQgupV2b+v2sPeOmkDRhzjAp7HC
yRcUnTmQM32LMO57oFZ7K9DJDMwxbHCAE7N8f2Uge2VLpIUjz3GskdjPcLBqu4qH
938dL3GMvKglo18VRSKiTkwxjxnP8HA8+0ifRv7cIQKBgQD7irdN/lYXdlhKAXwt
eQ5Tua3/Zi3Xia0dgOEzGxfVT1L6rVjVa149/v7x26p0evt0c9gkcHlBgUDEH7yz
sEDY/KwIo9Wg2Bpe1iUNcO/BvZc0mc3Cu9D8agxNZ5eHq+a36gTKR/QD1dp+48zK
6yimXxEfUYy2u8TBLm2Jk3hx4QKBgQDl0RA+L/SP8d2lB4vPDSSFhMYnwaH4njbq
/S8evNFcyKba9HaA8Sbng/msI/k8rnyul5GqFVooVl9TfPa93ATvagzijBvvqZ6R
j2+41hEAxmZ2257DKYpQ4JM2mRl9pOFQrc1GRbUfSOJAekj2O/2ojnbHHUcxg0dS
yliskq6BJwKBgGxVDKcBb5CBPnr48sMezMXQRRimp/2Y5L69H8AD3hrXI/SkLYsU
x6zJooEFSv8JbDx2G9NtwTst8HfG910n/nW1NF4wOTQhfhH0BlcomYmGHpXf25cP
jmz3Oz8m60LaDO6OUevQW04/ju9xKmUGLCai8NvdIk4cxhsw5KoIoinhAoGBALDu
vCqkkQ0hkRs1LBZEcBG7nzOMiD740B8qvdRUWnusn4mDHJk5EFK98MLvDzwAuk1Q
s/zWY4satFl6pByX/9SzOShR5lAlrscyPzl21bBbDxgDDcADg1GxFKW8STvKbQ3I
QXoQwNlNK6OogfPRTAExbZDuoZklEQxUbOCwLVmRAoGACIi7DWWZr9Ea4ag82yYv
CBdlKSshb2RYGdbYm9dzeqqv1/xtwtFYMOuLilL3qp9erBAsc6Kdz+n3v1+2p7Ho
3/upcFI5QbhGHCO5J92fdDYh6rsAG4cwC5LlAC5WgP3bL9+QVKKcYm8/mmglLIu3
XIvgj23tDAmXtVRh0MsgOH4=
-----END PRIVATE KEY-----
"""


@pytest.fixture()
def jwt_token(jwt_signing_key: str) -> str:
    payload = {
        "sub": "user123",
        "aud": "mrok-audience",
        "iss": "http:/test.mrok/",
        "exp": datetime.now(UTC) + timedelta(minutes=30),
    }
    return jwt.encode(payload, jwt_signing_key, algorithm="RS256", headers={"kid": "test-key"})


@pytest.fixture()
def jwks_json() -> dict:
    return {
        "keys": [
            {
                "kty": "RSA",
                "use": "sig",
                "alg": "RS256",
                "kid": "test-key",
                "n": (
                    "4dCCLjFbsV0qtZ3Zjtd0S6ar_tzXcodGhD_C89_80HspYmAn1vBzKD_"
                    "L4wvX7RinRByxuHqFLOjRfvHWIdzEbQ4pJdKB_yfv-gxUt86hvjTWjF"
                    "KU4JirK59cOTDfzT0MT6fw43OJwHAKIlf3NMOX37eQi6C5C7eYi_Mtp"
                    "k7kdFUYWen0voUIkDHEgWo7a_Izy4XpvbKSedzecV_zkDU5D-9drPak"
                    "sAi3fEqBczRWGptpByrPp3zG15pND12y402_UpOqTmuBMuTmVP__KBc"
                    "jJxJbbkVB9efFrJKcUrMcGaLDzrKxV3BHauQxz0KaXOR-iI8Hr1a_FW"
                    "QmfctA4am6Rw"
                ),
                "e": "AQAB",
            }
        ]
    }


@pytest.fixture()
def openid_config() -> dict:
    return {"issuer": "http:/test.mrok/", "jwks_uri": "http://example.com/jwks.json"}


@pytest.fixture()
def fastapi_app(settings_factory: SettingsFactory) -> FastAPI:
    settings = settings_factory()
    from mrok.controller.app import setup_app

    app = setup_app()
    app.dependency_overrides[get_settings] = lambda: settings
    return app


@pytest.fixture()
async def app_lifespan_manager(fastapi_app: FastAPI) -> AsyncGenerator[LifespanManager, None]:
    async with LifespanManager(fastapi_app) as lifespan_manager:
        yield lifespan_manager


@pytest.fixture
async def api_client(
    fastapi_app: FastAPI,
    app_lifespan_manager: LifespanManager,
    settings_factory: SettingsFactory,
    httpx_mock: HTTPXMock,
    openid_config: dict,
    jwks_json: dict,
    jwt_token: str,
) -> AsyncGenerator[AsyncClient]:
    settings = settings_factory()
    httpx_mock.add_response(
        method="GET",
        url=settings.auth.openid_config_url,
        json=openid_config,
        is_reusable=True,
    )
    httpx_mock.add_response(
        method="GET",
        url="http://example.com/jwks.json",
        json=jwks_json,
        is_reusable=True,
    )

    async with AsyncClient(
        transport=ASGITransport(app=app_lifespan_manager.app),
        base_url=f"http://localhost/{fastapi_app.root_path.removeprefix('/')}/",
        headers={"Authorization": f"Bearer {jwt_token}"},
    ) as client:
        yield client



import shlex

import pytest
from pytest_mock import MockerFixture
from typer.testing import CliRunner

from mrok.cli import app


def test_run_asgi(
    mocker: MockerFixture,
):
    mocked_ziticorn = mocker.patch("mrok.cli.commands.agent.run.asgi.ziticorn.run")
    runner = CliRunner()

    # Run the command
    result = runner.invoke(
        app,
        shlex.split("agent run asgi my:app ins-1234-5678-0001.json -w 2 --reload"),
    )
    assert result.exit_code == 0
    mocked_ziticorn.assert_called_once_with(
        "my:app", "ins-1234-5678-0001.json", workers=2, reload=True
    )


@pytest.mark.parametrize(
    ("target_addr", "expected_target_addr"),
    [
        ("/path/to/unix.sock", "/path/to/unix.sock"),
        ("localhost:8000", ("localhost", 8000)),
        (":8000", ("127.0.0.1", 8000)),
    ],
)
def test_run_sidecar(
    mocker: MockerFixture,
    target_addr,
    expected_target_addr,
):
    mocked_sidecar = mocker.patch("mrok.cli.commands.agent.run.sidecar.sidecar.run")
    runner = CliRunner()

    # Run the command
    result = runner.invoke(
        app,
        shlex.split(
            f"agent run sidecar ins-1234-5678-0001.json {target_addr}  -w 2 --reload",
        ),
    )
    assert result.exit_code == 0
    mocked_sidecar.assert_called_once_with(
        "ins-1234-5678-0001.json", expected_target_addr, workers=2, reload=True
    )



import logging
from typing import Any

from mrok.ziti.api import TagsType, ZitiClientAPI, ZitiManagementAPI
from mrok.ziti.identities import enroll_proxy_identity

logger = logging.getLogger(__name__)


async def bootstrap_identity(
    mgmt_api: ZitiManagementAPI,
    client_api: ZitiClientAPI,
    identity_name: str,
    mode: str,
    forced: bool,
    tags: TagsType | None,
) -> tuple[str, dict[str, Any] | None]:
    logger.info(f"Bootstrapping '{identity_name}' identity...")

    identity_json = None
    existing_identity = await mgmt_api.search_identity(identity_name)

    if forced and existing_identity:
        logger.info(f"Deleting existing identity '{identity_name}' ({existing_identity['id']})")

        policy = await mgmt_api.search_router_policy(identity_name)
        if policy:
            await mgmt_api.delete_router_policy(policy["id"])
            logger.info(f"Deleted existing ERP '{policy['name']}' ({policy['id']})")

        await mgmt_api.delete_identity(existing_identity["id"])
        logger.info("Deleted existing identity")
        existing_identity = None

    if existing_identity:
        frontend_id = existing_identity["id"]
        logger.info(f"Identity '{identity_name}' ({frontend_id}) is already enrolled")
    else:
        frontend_id, identity_json = await enroll_proxy_identity(
            mgmt_api,
            client_api,
            identity_name,
            tags=tags,
        )
        logger.info(f"Identity '{identity_name}' ({frontend_id}) successfully enrolled")

    policy = await mgmt_api.search_router_policy(identity_name)
    if not policy:
        policy_id = await mgmt_api.create_router_policy(
            identity_name,
            frontend_id,
            tags=tags,
        )
        logger.info(f"Created ERP '{identity_name}' ({policy_id})")
    else:
        logger.info(f"Found ERP '{policy['name']}' ({policy['id']})")

    config_type_name = f"{mode}.proxy.v1"
    config_type = await mgmt_api.search_config_type(config_type_name)
    if config_type is None:
        config_type_id = await mgmt_api.create_config_type(config_type_name, tags=tags)
        logger.info(f"Created '{config_type_name}' ({config_type_id}) config type")
    else:
        logger.info(f"Found '{config_type_name}' ({config_type['id']}) config type")

    if config_type and existing_identity:
        logger.info(f"Identity '{identity_name}' was already bootstrapped")
    else:
        logger.info("Bootstrap completed")

    return frontend_id, identity_json



import typer

from mrok.cli.commands.admin import bootstrap
from mrok.cli.commands.admin.register import app as register_app
from mrok.cli.commands.admin.unregister import app as unregister_app

app = typer.Typer(help="mrok administrative commands.")
app.add_typer(register_app)
app.add_typer(unregister_app)
bootstrap.register(app)



import logging

from uvicorn.protocols.http.httptools_impl import HttpToolsProtocol


class MrokHttpToolsProtocol(HttpToolsProtocol):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.logger = logging.getLogger("mrok.proxy")
        self.access_logger = logging.getLogger("mrok.access")
        self.access_log = self.access_logger.hasHandlers()



import shlex
import tempfile

import pytest
from pytest_mock import MockerFixture
from typer import BadParameter
from typer.testing import CliRunner

from mrok.cli import app
from mrok.cli.commands.admin.register.extensions import do_register as do_register_extension
from mrok.cli.commands.admin.register.instances import do_register as do_register_instance
from mrok.cli.commands.admin.utils import parse_tags
from tests.conftest import SettingsFactory


def test_register_extension_command(
    mocker: MockerFixture,
    settings_factory: SettingsFactory,
):
    settings = settings_factory()
    mocker.patch("mrok.cli.main.get_settings", return_value=settings)
    mock_register_coro = mocker.MagicMock()
    mock_register = mocker.MagicMock(return_value=mock_register_coro)

    mocker.patch("mrok.cli.commands.admin.register.extensions.do_register", mock_register)
    mock_run = mocker.patch("mrok.cli.commands.admin.register.extensions.asyncio.run")
    runner = CliRunner()

    # Run the command
    result = runner.invoke(
        app,
        shlex.split("admin register extension EXT-1234-5678 --tag tag=tagvalue"),
    )
    assert result.exit_code == 0
    mock_run.assert_called_once_with(mock_register_coro)

    mock_register.assert_called_once_with(
        settings,
        "EXT-1234-5678",
        ["tag=tagvalue"],
    )


def test_register_extension_command_invalid_extension_id(
    mocker: MockerFixture,
    settings_factory: SettingsFactory,
):
    settings = settings_factory()
    mocker.patch("mrok.cli.main.get_settings", return_value=settings)
    mock_register_coro = mocker.MagicMock()
    mock_register = mocker.MagicMock(return_value=mock_register_coro)

    mocker.patch("mrok.cli.commands.admin.register.extensions.do_register", mock_register)
    mocker.patch("mrok.cli.commands.admin.register.extensions.asyncio.run")
    runner = CliRunner()

    # Run the command
    result = runner.invoke(
        app,
        shlex.split("admin register extension EXT-1234 --tag tag=tagvalue"),
    )
    assert result.exit_code == 2
    assert "Invalid value for 'EXTENSION_ID': ext_id must match EXT-xxxx-yyyy" in result.stderr


@pytest.mark.asyncio
async def test_do_register_extension(mocker: MockerFixture, settings_factory: SettingsFactory):
    settings = settings_factory()
    mocked_api = mocker.AsyncMock()
    mocked_api.__aenter__.return_value = mocked_api
    mocked_api_ctor = mocker.patch(
        "mrok.cli.commands.admin.register.extensions.ZitiManagementAPI",
        return_value=mocked_api,
    )
    mocked_register_service = mocker.patch(
        "mrok.cli.commands.admin.register.extensions.register_extension"
    )

    await do_register_extension(settings, "EXT-1234", ["tag=value"])
    mocked_register_service.assert_awaited_once_with(
        settings, mocked_api, "EXT-1234", tags={"tag": "value"}
    )
    mocked_api_ctor.assert_called_once_with(settings)


def test_register_instance_command(
    mocker: MockerFixture,
    settings_factory: SettingsFactory,
):
    settings = settings_factory()
    mocker.patch("mrok.cli.main.get_settings", return_value=settings)
    mock_register_coro = mocker.MagicMock()
    mock_register = mocker.MagicMock(return_value=mock_register_coro)

    mocker.patch("mrok.cli.commands.admin.register.instances.do_register", mock_register)
    mock_run = mocker.patch(
        "mrok.cli.commands.admin.register.instances.asyncio.run",
        return_value=({"id": "identity_id"}, {"identity": "json"}),
    )
    runner = CliRunner()
    with tempfile.NamedTemporaryFile() as f:
        # Run the command
        result = runner.invoke(
            app,
            shlex.split(
                "admin register instance EXT-1234-5678 "
                f"INS-1234-5678-0001 {f.name} --tag tag=tagvalue"
            ),
        )
        f.seek(0)
        json_data = f.read()
        assert json_data == b'{"identity": "json"}'
        assert result.exit_code == 0

    mock_run.assert_called_once_with(mock_register_coro)

    mock_register.assert_called_once_with(
        settings,
        "EXT-1234-5678",
        "INS-1234-5678-0001",
        ["tag=tagvalue"],
    )


def test_register_instance_command_invalid_extension_id(
    mocker: MockerFixture,
    settings_factory: SettingsFactory,
):
    settings = settings_factory()
    mocker.patch("mrok.cli.main.get_settings", return_value=settings)
    mock_register_coro = mocker.MagicMock()
    mock_register = mocker.MagicMock(return_value=mock_register_coro)

    mocker.patch("mrok.cli.commands.admin.register.instances.do_register", mock_register)
    mocker.patch("mrok.cli.commands.admin.register.instances.asyncio.run")
    runner = CliRunner()

    # Run the command
    result = runner.invoke(
        app,
        shlex.split(
            "admin register instance EXT-1234 INS-1234-0001 my-file.json --tag tag=tagvalue",
        ),
    )
    assert result.exit_code == 2
    assert "Invalid value for 'EXTENSION_ID': ext_id must match EXT-xxxx-yyyy" in result.stderr


@pytest.mark.asyncio
async def test_do_register_instance(mocker: MockerFixture, settings_factory: SettingsFactory):
    settings = settings_factory()
    mocked_mgmt_api = mocker.AsyncMock()
    mocked_mgmt_api.__aenter__.return_value = mocked_mgmt_api
    mocked_mgmt_api_ctor = mocker.patch(
        "mrok.cli.commands.admin.register.instances.ZitiManagementAPI",
        return_value=mocked_mgmt_api,
    )
    mocked_cli_api = mocker.AsyncMock()
    mocked_cli_api.__aenter__.return_value = mocked_cli_api
    mocked_cli_api_ctor = mocker.patch(
        "mrok.cli.commands.admin.register.instances.ZitiClientAPI",
        return_value=mocked_cli_api,
    )
    mocked_register_instance = mocker.patch(
        "mrok.cli.commands.admin.register.instances.register_instance",
    )

    await do_register_instance(settings, "EXT-1234", "INS-1234-0001", ["tag=value"])
    mocked_register_instance.assert_awaited_once_with(
        mocked_mgmt_api, mocked_cli_api, "EXT-1234", "INS-1234-0001", tags={"tag": "value"}
    )
    mocked_mgmt_api_ctor.assert_called_once_with(settings)
    mocked_cli_api_ctor.assert_called_once_with(settings)


def test_parse_tags():
    assert parse_tags(
        [
            "str=value",
            "true=true",
            "false=false",
            "null=",
        ]
    ) == {
        "str": "value",
        "true": True,
        "false": False,
        "null": None,
    }


@pytest.mark.parametrize("tags_list", [None, []])
def test_parse_tags_no_tags(tags_list):
    assert parse_tags(tags_list) is None


def test_parse_tags_invalid():
    with pytest.raises(BadParameter) as cv:
        parse_tags(["blabla"])
    assert str(cv.value) == "Invalid format 'blabla', expected key=value"



import json
import shlex
from pathlib import Path

import yaml
from pytest_mock import MockerFixture
from typer.testing import CliRunner

from mrok.cli import app


def test_openapi(mocker: MockerFixture):
    spec = {"test": "openapi"}
    mocker.patch("mrok.cli.commands.controller.openapi.generate_openapi_spec", return_value=spec)
    mocked_open = mocker.mock_open()
    mocker.patch("mrok.cli.commands.controller.openapi.open", mocked_open)

    runner = CliRunner()
    result = runner.invoke(app, shlex.split("controller openapi"))

    assert result.exit_code == 0
    mocked_open.assert_called_once_with(Path("mrok_openapi_spec.yml"), "w")
    written_data = "".join(call.args[0] for call in mocked_open().write.call_args_list)
    assert written_data == yaml.dump(spec, indent=2)


def test_openapi_custom_output(mocker: MockerFixture):
    spec = {"test": "openapi"}
    mocker.patch("mrok.cli.commands.controller.openapi.generate_openapi_spec", return_value=spec)
    mocked_open = mocker.mock_open()
    mocker.patch("mrok.cli.commands.controller.openapi.open", mocked_open)

    runner = CliRunner()
    result = runner.invoke(app, shlex.split("controller openapi -o openapi.json -f json"))

    assert result.exit_code == 0
    mocked_open.assert_called_once_with(Path("openapi.json"), "w")
    written_data = "".join(call.args[0] for call in mocked_open().write.call_args_list)
    assert written_data == json.dumps(spec, indent=2)



import logging
from typing import Annotated

from fastapi import APIRouter, Body, HTTPException, status

from mrok.controller.dependencies import AppSettings, ZitiClientAPI, ZitiManagementAPI
from mrok.controller.openapi import examples
from mrok.controller.pagination import LimitOffsetPage, paginate
from mrok.controller.schemas import ExtensionCreate, ExtensionRead, InstanceCreate, InstanceRead
from mrok.ziti.constants import MROK_SERVICE_TAG_NAME
from mrok.ziti.errors import (
    ConfigTypeNotFoundError,
    ProxyIdentityNotFoundError,
    ServiceAlreadyRegisteredError,
    ServiceNotFoundError,
)
from mrok.ziti.identities import register_instance, unregister_instance
from mrok.ziti.services import register_extension, unregister_extension

logger = logging.getLogger("mrok.controller")

router = APIRouter()


async def fetch_extension_or_404(mgmt_api: ZitiManagementAPI, id_or_extension_id: str):
    service = await mgmt_api.search_service(id_or_extension_id)
    if not service:
        raise HTTPException(
            status_code=status.HTTP_404_NOT_FOUND,
        )
    return service


async def fetch_instance_or_404(
    mgmt_api: ZitiManagementAPI, id_or_extension_id: str, id_or_instance_id: str
):
    service = await fetch_extension_or_404(mgmt_api, id_or_extension_id)
    if id_or_instance_id.startswith("INS-"):
        id_or_name = f"{id_or_instance_id}.{service['name']}"
    else:
        id_or_name = id_or_instance_id
    identity = await mgmt_api.search_identity(id_or_name)
    if not identity:
        raise HTTPException(
            status_code=status.HTTP_404_NOT_FOUND,
        )
    return identity


@router.post(
    "",
    response_model=ExtensionRead,
    responses={
        201: {
            "description": "Extension",
            "content": {
                "application/json": {
                    "example": examples.EXTENSION_RESPONSE,
                }
            },
        },
    },
    status_code=status.HTTP_201_CREATED,
    tags=["Extensions"],
)
async def create_extension(
    settings: AppSettings,
    mgmt_api: ZitiManagementAPI,
    data: Annotated[
        ExtensionCreate,
        Body(
            openapi_examples={
                "create_extension": {
                    "summary": "Create an Extension",
                    "description": ("Create a new Extension."),
                    "value": {
                        "extension": {"id": "EXT-1234-5678"},
                        "tags": {"account": "ACC-5555-3333"},
                    },
                }
            }
        ),
    ],
):
    try:
        service = await register_extension(settings, mgmt_api, data.extension.id, data.tags)
        return ExtensionRead(
            id=service["id"],
            name=service["name"],
            tags=service["tags"],
        )
    except (ProxyIdentityNotFoundError, ConfigTypeNotFoundError) as e:
        raise HTTPException(
            status_code=status.HTTP_400_BAD_REQUEST,
            detail=f"OpenZiti not configured properly: {e}",
        ) from e
    except ServiceAlreadyRegisteredError as e:
        raise HTTPException(
            status_code=status.HTTP_400_BAD_REQUEST,
            detail=str(e),
        ) from e


@router.get(
    "/{id_or_extension_id}",
    response_model=ExtensionRead,
    responses={
        200: {
            "description": "Extension",
            "content": {"application/json": {"example": examples.EXTENSION_RESPONSE}},
        },
    },
    dependencies=[],
    tags=["Extensions"],
)
async def get_extension_by_id_or_extension_id(
    mgmt_api: ZitiManagementAPI,
    id_or_extension_id: str,
):
    return ExtensionRead(**(await fetch_extension_or_404(mgmt_api, id_or_extension_id)))


@router.delete(
    "/{id_or_extension_id}",
    status_code=status.HTTP_204_NO_CONTENT,
    tags=["Extensions"],
)
async def delete_instance_by_id_or_extension_id(
    settings: AppSettings,
    mgmt_api: ZitiManagementAPI,
    id_or_extension_id: str,
):
    try:
        await unregister_extension(settings, mgmt_api, id_or_extension_id)
    except ServiceNotFoundError:
        raise HTTPException(
            status_code=status.HTTP_404_NOT_FOUND,
        )


@router.get(
    "",
    response_model=LimitOffsetPage[ExtensionRead],
    responses={
        200: {
            "description": "List of Extensions",
            "content": {
                "application/json": {
                    "example": {
                        "data": [examples.EXTENSION_RESPONSE],
                        "$meta": {
                            "pagination": {
                                "total": 1,
                                "limit": 10,
                                "offset": 0,
                            },
                        },
                    },
                },
            },
        },
    },
    tags=["Extensions"],
)
async def get_extensions(
    mgmt_api: ZitiManagementAPI,
):
    return await paginate(mgmt_api, "/services", ExtensionRead)


@router.post(
    "/{id_or_extension_id}/instances",
    response_model=InstanceRead,
    responses={
        201: {
            "description": "Instance",
            "content": {
                "application/json": {
                    "example": examples.INSTANCE_CREATE_RESPONSE,
                }
            },
        },
    },
    status_code=status.HTTP_201_CREATED,
    tags=["Instances"],
)
async def create_extension_instances(
    mgmt_api: ZitiManagementAPI,
    client_api: ZitiClientAPI,
    id_or_extension_id: str,
    data: Annotated[
        InstanceCreate,
        Body(
            openapi_examples={
                "create_extension": {
                    "summary": "Create an Extension Instance",
                    "description": ("Create a new Instance of an Extension."),
                    "value": {
                        "instance": {"id": "INS-1234-5678-0001"},
                        "tags": {"account": "ACC-5555-3333"},
                    },
                }
            }
        ),
    ],
):
    service = await fetch_extension_or_404(mgmt_api, id_or_extension_id)
    identity, identity_file = await register_instance(
        mgmt_api, client_api, service["name"], data.instance.id, data.tags
    )
    return InstanceRead(
        id=identity["id"],
        name=identity["name"],
        identity=identity_file,
        tags=identity["tags"],
    )


@router.get(
    "/{id_or_extension_id}/instances",
    response_model=LimitOffsetPage[InstanceRead],
    responses={
        200: {
            "description": "List of Instances.",
            "content": {
                "application/json": {
                    "example": {
                        "data": [examples.INSTANCE_RESPONSE],
                        "$meta": {
                            "pagination": {
                                "total": 1,
                                "limit": 10,
                                "offset": 0,
                            },
                        },
                    },
                },
            },
        },
    },
    dependencies=[],
    tags=["Instances"],
)
async def list_extension_instances(
    mgmt_api: ZitiManagementAPI,
    id_or_extension_id: str,
):
    service = await fetch_extension_or_404(mgmt_api, id_or_extension_id)
    return await paginate(
        mgmt_api,
        "/identities",
        InstanceRead,
        {"filter": f'tags.{MROK_SERVICE_TAG_NAME} = "{service["name"]}"'},
    )


@router.get(
    "/{id_or_extension_id}/instances/{id_or_instance_id}",
    response_model=InstanceRead,
    responses={
        200: {
            "description": "Instance",
            "content": {"application/json": {"example": examples.INSTANCE_RESPONSE}},
        },
    },
    dependencies=[],
    tags=["Instances"],
)
async def get_instance_by_id_or_instance_id(
    mgmt_api: ZitiManagementAPI,
    id_or_extension_id: str,
    id_or_instance_id: str,
):
    identity = await fetch_instance_or_404(mgmt_api, id_or_extension_id, id_or_instance_id)
    return InstanceRead(
        id=identity["id"],
        name=identity["name"],
        tags=identity["tags"],
    )


@router.delete(
    "/{id_or_extension_id}/instances/{id_or_instance_id}",
    status_code=status.HTTP_204_NO_CONTENT,
    tags=["Instances"],
)
async def delete_instance_by_id_or_instance_id(
    mgmt_api: ZitiManagementAPI,
    id_or_extension_id: str,
    id_or_instance_id: str,
):
    identity = await fetch_instance_or_404(mgmt_api, id_or_extension_id, id_or_instance_id)
    instance_id, extension_id = identity["name"].split(".")
    await unregister_instance(mgmt_api, extension_id, instance_id)



import logging.config

from mrok.conf import Settings


def get_logging_config(settings: Settings, cli_mode: bool = False) -> dict:
    log_level = "DEBUG" if settings.logging.debug else "INFO"
    handler = "rich" if settings.logging.rich else "console"

    if cli_mode:
        mrok_handler = "cli"
    else:
        mrok_handler = handler

    logging_config = {
        "version": 1,
        "disable_existing_loggers": False,
        "formatters": {
            "verbose": {
                "format": "{asctime} {name} {levelname} (pid: {process}) {message}",
                "style": "{",
            },
            "rich": {
                "format": "{name} {message}",
                "style": "{",
            },
            "plain": {"format": "%(message)s"},
        },
        "handlers": {
            "console": {
                "class": "logging.StreamHandler",
                "formatter": "verbose",
                "stream": "ext://sys.stderr",
            },
            "cli": {
                "class": "logging.StreamHandler",
                "formatter": "plain",
                "stream": "ext://sys.stdout",
            },
            "rich": {
                "class": "rich.logging.RichHandler",
                "level": log_level,
                "formatter": "rich",
                "log_time_format": lambda x: x.strftime("%Y-%m-%d %H:%M:%S.%f")[:-3],
                "rich_tracebacks": True,
            },
        },
        "root": {
            "handlers": ["rich"],
            "level": "WARNING",
        },
        "loggers": {
            "gunicorn.access": {
                "handlers": [handler],
                "level": log_level,
                "propagate": False,
            },
            "gunicorn.error": {
                "handlers": [handler],
                "level": log_level,
                "propagate": False,
            },
            "mrok": {
                "handlers": [mrok_handler],
                "level": log_level,
                "propagate": False,
            },
        },
    }

    return logging_config


def setup_logging(settings: Settings, cli_mode: bool = False) -> None:
    logging_config = get_logging_config(settings, cli_mode)
    logging.config.dictConfig(logging_config)



from pytest_mock import MockerFixture

from mrok.http.lifespan import MrokLifespan


def test_lifespan(mocker: MockerFixture):
    mocked_config = mocker.MagicMock(loaded=True)

    lifespan = MrokLifespan(mocked_config)
    assert lifespan.config == mocked_config
    assert lifespan.logger.name == "mrok.proxy"



import asyncio
import logging
from collections.abc import Awaitable, Callable
from pathlib import Path
from typing import Any

from mrok.http.forwarder import ForwardAppBase

logger = logging.getLogger("mrok.proxy")

Scope = dict[str, Any]
ASGIReceive = Callable[[], Awaitable[dict[str, Any]]]
ASGISend = Callable[[dict[str, Any]], Awaitable[None]]


class ForwardApp(ForwardAppBase):
    def __init__(
        self, target_address: str | Path | tuple[str, int], read_chunk_size: int = 65536
    ) -> None:
        super().__init__(read_chunk_size=read_chunk_size)
        self._target_address = target_address

    async def select_backend(
        self,
        scope: Scope,
        headers: dict[str, str],
    ) -> tuple[asyncio.StreamReader, asyncio.StreamWriter] | tuple[None, None]:
        if isinstance(self._target_address, tuple):
            return await asyncio.open_connection(*self._target_address)
        return await asyncio.open_unix_connection(str(self._target_address))



import logging
import socket

from uvicorn import server

server.logger = logging.getLogger("mrok.proxy")


class MrokServer(server.Server):
    async def serve(self, sockets: list[socket.socket] | None = None) -> None:
        if not sockets:
            sockets = [self.config.bind_socket()]
        with self.capture_signals():
            await self._serve(sockets)



import inspect
import sys

import typer
from pyfiglet import Figlet
from rich.text import Text
from typer.core import TyperGroup

from mrok.cli import commands
from mrok.cli.rich import get_console
from mrok.conf import get_settings
from mrok.logging import setup_logging


def gradient(start_hex, end_hex, num_samples=10):
    start_rgb = tuple(int(start_hex[i : i + 2], 16) for i in range(1, 6, 2))
    end_rgb = tuple(int(end_hex[i : i + 2], 16) for i in range(1, 6, 2))
    gradient_colors = [start_hex]
    for sample in range(1, num_samples):
        red = int(start_rgb[0] + (float(sample) / (num_samples - 1)) * (end_rgb[0] - start_rgb[0]))
        green = int(
            start_rgb[1] + (float(sample) / (num_samples - 1)) * (end_rgb[1] - start_rgb[1])
        )
        blue = int(start_rgb[2] + (float(sample) / (num_samples - 1)) * (end_rgb[2] - start_rgb[2]))
        gradient_colors.append(f"#{red:02X}{green:02X}{blue:02X}")

    return gradient_colors


def show_banner():
    program_name = "mrok"
    figlet = Figlet("georgia11")

    banner_text = figlet.renderText(program_name)

    banner_lines = [Text(line) for line in banner_text.splitlines()]
    max_line_length = max(len(line) for line in banner_lines)
    half_length = max_line_length // 2

    colors = gradient("#00C9CD", "#472AFF", half_length) + gradient(
        "#472AFF", "#392D9C", half_length + 1
    )
    console = get_console()

    for line in banner_lines:
        colored_line = Text()
        for i in range(len(line)):
            char = line[i : i + 1]
            char.stylize(colors[i])
            colored_line = Text.assemble(colored_line, char)
        console.print(colored_line)


class MrokGroup(TyperGroup):
    def get_help(self, ctx):
        show_banner()
        return super().get_help(ctx)

    def invoke(self, ctx):
        show_banner()
        return super().invoke(ctx)


app = typer.Typer(
    help="SoftwareOne Marketplace Extension Channel CLI",
    add_completion=True,
    rich_markup_mode="rich",
    cls=MrokGroup,
)

err_console = get_console(stderr=True)

for name, module in inspect.getmembers(commands):
    if not inspect.ismodule(module):
        continue

    if hasattr(module, "command"):  # pragma: no branch
        app.command(name=name.replace("_", "-"))(module.command)
    elif hasattr(module, "app"):  # pragma: no branch
        app.add_typer(module.app, name=name.replace("_", "-"))


@app.callback()
def main(
    ctx: typer.Context,
):
    settings = get_settings()
    setup_logging(settings, cli_mode=True)
    ctx.obj = settings


def run():
    try:
        app()
    except Exception as e:
        err_console.print(f"[bold red]Error:[/bold red] {e}")
        sys.exit(-1)



from typing import Any
from urllib.parse import quote

import pytest
from pytest_httpx import HTTPXMock
from pytest_mock import MockerFixture

from mrok.ziti.api import (
    TagsType,
    ZitiAuthError,
    ZitiBadRequestError,
    ZitiClientAPI,
    ZitiIdentityAuthContext,
    ZitiManagementAPI,
)
from mrok.ziti.constants import MROK_VERSION_TAG_NAME
from tests.conftest import SettingsFactory


@pytest.mark.asyncio
@pytest.mark.parametrize("tags", [None, {"foo": "bar"}])
async def test_create(
    settings_factory: SettingsFactory,
    httpx_mock: HTTPXMock,
    tags: TagsType | None,
):
    settings = settings_factory()
    expected_body = {
        "test": "payload",
        "tags": {MROK_VERSION_TAG_NAME: "0.0.0.dev0", **(tags or {})},
    }
    httpx_mock.add_response(
        method="POST",
        url=f"{settings.ziti.api.management}/edge/management/v1/services",
        match_json=expected_body,
        json={"data": {"id": "service123"}},
    )
    async with ZitiManagementAPI(settings) as api:
        result = await api.create("/services", {"test": "payload"}, tags)
        assert result == "service123"


@pytest.mark.asyncio
async def test_create_bad_request(
    settings_factory: SettingsFactory,
    httpx_mock: HTTPXMock,
    ziti_bad_request_error: dict[str, Any],
):
    settings = settings_factory()
    httpx_mock.add_response(
        method="POST",
        url=f"{settings.ziti.api.management}/edge/management/v1/services",
        status_code=400,
        json=ziti_bad_request_error,
    )
    async with ZitiManagementAPI(settings) as api:
        with pytest.raises(ZitiBadRequestError) as cv:
            await api.create("/services", {"test": "payload"}, None)
    assert str(cv.value) == "error_code - error_message (field_name: field_error)"


@pytest.mark.asyncio
async def test_get(
    settings_factory: SettingsFactory,
    httpx_mock: HTTPXMock,
):
    settings = settings_factory()
    httpx_mock.add_response(
        method="GET",
        url=f"{settings.ziti.api.management}/edge/management/v1/services/service123",
        json={"data": {"id": "service123", "name": "svc"}},
    )
    async with ZitiManagementAPI(settings) as api:
        result = await api.get("/services", "service123")
    assert result["id"] == "service123"
    assert result["name"] == "svc"


@pytest.mark.asyncio
async def test_delete(
    settings_factory: SettingsFactory,
    httpx_mock: HTTPXMock,
):
    settings = settings_factory()
    httpx_mock.add_response(
        method="DELETE",
        url=f"{settings.ziti.api.management}/edge/management/v1/services/service123",
        json={"data": {"deleted": True}},
    )
    async with ZitiManagementAPI(settings) as api:
        await api.delete("/services", "service123")


@pytest.mark.asyncio
async def test_search_by_id_or_name(settings_factory: SettingsFactory, httpx_mock: HTTPXMock):
    settings = settings_factory()
    query = quote(f'(id="svc" or name="svc") and tags.{MROK_VERSION_TAG_NAME} != null')
    httpx_mock.add_response(
        method="GET",
        url=f"{settings.ziti.api.management}/edge/management/v1/services?filter={query}",
        json={
            "meta": {"pagination": {"totalCount": 1}},
            "data": [{"id": "svc1", "name": "svc"}],
        },
    )
    async with ZitiManagementAPI(settings) as api:
        result = await api.search_by_id_or_name("/services", "svc")
    assert result is not None
    assert result["id"] == "svc1"


@pytest.mark.asyncio
async def test_search_by_id_or_name_no_results(
    settings_factory: SettingsFactory, httpx_mock: HTTPXMock
):
    settings = settings_factory()
    query = quote(f'(id="svc" or name="svc") and tags.{MROK_VERSION_TAG_NAME} != null')
    httpx_mock.add_response(
        method="GET",
        url=f"{settings.ziti.api.management}/edge/management/v1/services?filter={query}",
        json={
            "meta": {"pagination": {"totalCount": 0}},
            "data": [],
        },
    )
    async with ZitiManagementAPI(settings) as api:
        result = await api.search_by_id_or_name("/services", "svc")
    assert result is None


@pytest.mark.asyncio
async def test_search_by_id_or_name_bad_request(
    settings_factory: SettingsFactory, httpx_mock: HTTPXMock, ziti_bad_request_error: dict[str, Any]
):
    settings = settings_factory()
    query = quote(f'(id="svc" or name="svc") and tags.{MROK_VERSION_TAG_NAME} != null')
    httpx_mock.add_response(
        method="GET",
        url=f"{settings.ziti.api.management}/edge/management/v1/services?filter={query}",
        status_code=400,
        json=ziti_bad_request_error,
    )
    async with ZitiManagementAPI(settings) as api:
        with pytest.raises(ZitiBadRequestError) as cv:
            await api.search_by_id_or_name("/services", "svc")
    assert str(cv.value) == "error_code - error_message (field_name: field_error)"


@pytest.mark.asyncio
async def test_collection_iterator(
    settings_factory: SettingsFactory,
    httpx_mock: HTTPXMock,
):
    settings = settings_factory()
    httpx_mock.add_response(
        method="GET",
        url=f"{settings.ziti.api.management}/edge/management/v1/services?limit=5&offset=0",
        json={
            "meta": {"pagination": {"totalCount": 10, "limit": 5, "offset": 0}},
            "data": [{"id": f"svc{i}", "name": "svc"} for i in range(5)],
        },
    )
    httpx_mock.add_response(
        method="GET",
        url=f"{settings.ziti.api.management}/edge/management/v1/services?limit=5&offset=5",
        json={
            "meta": {"pagination": {"totalCount": 10, "limit": 5, "offset": 5}},
            "data": [{"id": f"svc{5 + i}", "name": "svc"} for i in range(5)],
        },
    )
    async with ZitiManagementAPI(settings) as api:
        results = [item async for item in api.collection_iterator("/services")]
    assert len(results) == 10


@pytest.mark.asyncio
@pytest.mark.parametrize(
    ("method_name", "endpoint"),
    [
        ("get_service", "/services"),
        ("get_identity", "/identities"),
    ],
)
async def test_get_methods(
    mocker: MockerFixture, settings_factory: SettingsFactory, method_name: str, endpoint: str
):
    settings = settings_factory()
    mocked_get = mocker.patch.object(ZitiManagementAPI, "get", return_value={"returned": "object"})
    async with ZitiManagementAPI(settings) as api:
        method = getattr(api, method_name)
        result = await method("id")
    assert result == {"returned": "object"}
    mocked_get.assert_awaited_once_with(endpoint, "id")


@pytest.mark.asyncio
@pytest.mark.parametrize(
    ("method_name", "endpoint"),
    [
        ("search_service_router_policy", "/service-edge-router-policies"),
        ("search_router_policy", "/edge-router-policies"),
        ("search_service", "/services"),
        ("search_identity", "/identities"),
        ("search_config_type", "/config-types"),
        ("search_service_policy", "/service-policies"),
        ("search_config", "/configs"),
    ],
)
async def test_search_methods(
    mocker: MockerFixture, settings_factory: SettingsFactory, method_name: str, endpoint: str
):
    settings = settings_factory()
    mocked_search = mocker.patch.object(
        ZitiManagementAPI, "search_by_id_or_name", return_value={"returned": "object"}
    )
    async with ZitiManagementAPI(settings) as api:
        method = getattr(api, method_name)
        result = await method("id")
    assert result == {"returned": "object"}
    mocked_search.assert_awaited_once_with(endpoint, "id")


@pytest.mark.asyncio
@pytest.mark.parametrize(
    ("method_name", "endpoint"),
    [
        ("delete_config", "/configs"),
        ("delete_service_policy", "/service-policies"),
        ("delete_service_router_policy", "/service-edge-router-policies"),
        ("delete_router_policy", "/edge-router-policies"),
        ("delete_service", "/services"),
        ("delete_identity", "/identities"),
    ],
)
async def test_delete_methods(
    mocker: MockerFixture, settings_factory: SettingsFactory, method_name: str, endpoint: str
):
    settings = settings_factory()
    mocked_deleted = mocker.patch.object(ZitiManagementAPI, "delete")
    async with ZitiManagementAPI(settings) as api:
        method = getattr(api, method_name)
        await method("id")

    mocked_deleted.assert_awaited_once_with(endpoint, "id")


@pytest.mark.asyncio
@pytest.mark.parametrize(
    ("method_name", "endpoint"),
    [
        ("services", "/services"),
        ("identities", "/identities"),
    ],
)
async def test_iterator_methods(
    mocker: MockerFixture, settings_factory: SettingsFactory, method_name: str, endpoint: str
):
    settings = settings_factory()
    mocked_awaitable = mocker.AsyncMock()
    mocked_collection_iterator = mocker.patch.object(
        ZitiManagementAPI, "collection_iterator", return_value=mocked_awaitable
    )
    async with ZitiManagementAPI(settings) as api:
        method = getattr(api, method_name)
        awaitable = method()

    mocked_collection_iterator.assert_called_once_with(endpoint)
    assert awaitable == mocked_awaitable


@pytest.mark.asyncio
async def test_create_config(mocker: MockerFixture, settings_factory: SettingsFactory):
    settings = settings_factory()
    mocked_create = mocker.patch.object(ZitiManagementAPI, "create", return_value="id")
    async with ZitiManagementAPI(settings) as api:
        config_id = await api.create_config("my-config", "my-config-type-id")
    assert config_id == "id"
    mocked_create.assert_awaited_once_with(
        "/configs",
        {
            "configTypeId": "my-config-type-id",
            "name": "my-config",
            "data": {
                "auth_scheme": "none",
                "basic_auth": None,
                "interstitial": True,
                "oauth": None,
            },
        },
        None,
    )


@pytest.mark.asyncio
async def test_create_config_type(mocker: MockerFixture, settings_factory: SettingsFactory):
    settings = settings_factory()
    mocked_create = mocker.patch.object(ZitiManagementAPI, "create", return_value="id")
    async with ZitiManagementAPI(settings) as api:
        config_type_id = await api.create_config_type("my-config-type")
    assert config_type_id == "id"
    mocked_create.assert_awaited_once_with(
        "/config-types",
        {
            "name": "my-config-type",
            "schema": {},
        },
        None,
    )


@pytest.mark.asyncio
async def test_create_service(mocker: MockerFixture, settings_factory: SettingsFactory):
    settings = settings_factory()
    mocked_create = mocker.patch.object(ZitiManagementAPI, "create", return_value="id")
    async with ZitiManagementAPI(settings) as api:
        service_id = await api.create_service("my-service", "my-config-id")
    assert service_id == "id"
    mocked_create.assert_awaited_once_with(
        "/services",
        {
            "name": "my-service",
            "configs": ["my-config-id"],
            "encryptionRequired": True,
        },
        None,
    )


@pytest.mark.asyncio
async def test_create_service_router_policy(
    mocker: MockerFixture, settings_factory: SettingsFactory
):
    settings = settings_factory()
    mocked_create = mocker.patch.object(ZitiManagementAPI, "create", return_value="id")
    async with ZitiManagementAPI(settings) as api:
        policy_id = await api.create_service_router_policy("my-policy", "my-service")
    assert policy_id == "id"
    mocked_create.assert_awaited_once_with(
        "/service-edge-router-policies",
        {
            "name": "my-policy",
            "edgeRouterRoles": ["#all"],
            "serviceRoles": ["@my-service"],
            "semantic": "AllOf",
        },
        None,
    )


@pytest.mark.asyncio
async def test_create_router_policy(mocker: MockerFixture, settings_factory: SettingsFactory):
    settings = settings_factory()
    mocked_create = mocker.patch.object(ZitiManagementAPI, "create", return_value="id")
    async with ZitiManagementAPI(settings) as api:
        policy_id = await api.create_router_policy("my-policy", "my-identity")
    assert policy_id == "id"
    mocked_create.assert_awaited_once_with(
        "/edge-router-policies",
        {
            "name": "my-policy",
            "edgeRouterRoles": ["#all"],
            "identityRoles": ["@my-identity"],
            "semantic": "AllOf",
        },
        None,
    )


@pytest.mark.asyncio
async def test_create_dial_service_policy(mocker: MockerFixture, settings_factory: SettingsFactory):
    settings = settings_factory()
    mocked_create = mocker.patch.object(ZitiManagementAPI, "create", return_value="id")
    async with ZitiManagementAPI(settings) as api:
        policy_id = await api.create_dial_service_policy("my-policy", "my-service", "my-identity")
    assert policy_id == "id"
    mocked_create.assert_awaited_once_with(
        "/service-policies",
        {
            "name": "my-policy",
            "type": "Dial",
            "serviceRoles": ["@my-service"],
            "identityRoles": ["@my-identity"],
            "semantic": "AllOf",
        },
        None,
    )


@pytest.mark.asyncio
async def test_create_bind_service_policy(mocker: MockerFixture, settings_factory: SettingsFactory):
    settings = settings_factory()
    mocked_create = mocker.patch.object(ZitiManagementAPI, "create", return_value="id")
    async with ZitiManagementAPI(settings) as api:
        policy_id = await api.create_bind_service_policy("my-policy", "my-service", "my-identity")
    assert policy_id == "id"
    mocked_create.assert_awaited_once_with(
        "/service-policies",
        {
            "name": "my-policy",
            "type": "Bind",
            "serviceRoles": ["@my-service"],
            "identityRoles": ["@my-identity"],
            "semantic": "AllOf",
        },
        None,
    )


@pytest.mark.asyncio
async def test_create_user_identity(mocker: MockerFixture, settings_factory: SettingsFactory):
    settings = settings_factory()
    mocked_create = mocker.patch.object(ZitiManagementAPI, "create", return_value="id")
    async with ZitiManagementAPI(settings) as api:
        policy_id = await api.create_user_identity("my-identity")
    assert policy_id == "id"
    mocked_create.assert_awaited_once_with(
        "/identities",
        {
            "name": "my-identity",
            "type": "User",
            "isAdmin": False,
            "enrollment": {"ott": True},
        },
        None,
    )


@pytest.mark.asyncio
async def test_create_device_identity(mocker: MockerFixture, settings_factory: SettingsFactory):
    settings = settings_factory()
    mocked_create = mocker.patch.object(ZitiManagementAPI, "create", return_value="id")
    async with ZitiManagementAPI(settings) as api:
        policy_id = await api.create_device_identity("my-identity")
    assert policy_id == "id"
    mocked_create.assert_awaited_once_with(
        "/identities",
        {
            "name": "my-identity",
            "type": "Device",
            "isAdmin": False,
            "enrollment": {"ott": True},
        },
        None,
    )


@pytest.mark.asyncio
async def test_fetch_ca_certs(
    settings_factory: SettingsFactory,
    httpx_mock: HTTPXMock,
):
    settings = settings_factory()
    httpx_mock.add_response(
        method="GET",
        url=f"{settings.ziti.api.management}/edge/management/v1/.well-known/est/cacerts",
        text="-----BEGIN CERTIFICATE----\n...\n-----END CERTIFICATE-----",
    )
    async with ZitiManagementAPI(settings) as api:
        result = await api.fetch_ca_certificates()
    assert result == "-----BEGIN CERTIFICATE----\n...\n-----END CERTIFICATE-----"


@pytest.mark.asyncio
async def test_password_auth(
    settings_factory: SettingsFactory,
    httpx_mock: HTTPXMock,
):
    settings = settings_factory()
    httpx_mock.add_response(
        method="GET",
        url=f"{settings.ziti.api.management}/edge/management/v1/services/service123",
        status_code=401,
    )
    httpx_mock.add_response(
        method="POST",
        url=f"{settings.ziti.api.management}/edge/management/v1/authenticate?method=password",
        match_json={
            "username": settings.ziti.auth.username,
            "password": settings.ziti.auth.password,
        },
        status_code=200,
        json={"data": {"token": "auth-token"}},
    )
    httpx_mock.add_response(
        method="GET",
        url=f"{settings.ziti.api.management}/edge/management/v1/services/service123",
        match_headers={"zt-session": "auth-token"},
        json={"data": {"id": "service123", "name": "svc"}},
    )
    async with ZitiManagementAPI(settings) as api:
        result = await api.get("/services", "service123")
    assert result["id"] == "service123"
    assert result["name"] == "svc"


def test_zitiidentityauthcontext(mocker: MockerFixture, ziti_identity_file: str):
    mocked_ssl_context = mocker.MagicMock()
    mocked_create_ssl_ctx = mocker.patch(
        "mrok.ziti.api.ssl.create_default_context", return_value=mocked_ssl_context
    )
    ZitiIdentityAuthContext(ziti_identity_file)
    mocked_create_ssl_ctx.assert_called_once()
    mocked_ssl_context.load_cert_chain.assert_called_once()


@pytest.mark.asyncio
async def test_identity_auth(
    mocker: MockerFixture,
    settings_factory: SettingsFactory,
    httpx_mock: HTTPXMock,
):
    settings = settings_factory(
        ziti={
            "api": {"management": "https://ziti.example.com"},
            "read_timeout": 10,
            "ssl_verify": True,
            "auth": {"identity": "my-identity.json"},
        }
    )
    mocked_ctx = mocker.patch("mrok.ziti.api.ZitiIdentityAuthContext")
    httpx_mock.add_response(
        method="GET",
        url=f"{settings.ziti.api.management}/edge/management/v1/services/service123",
        status_code=401,
    )
    httpx_mock.add_response(
        method="POST",
        url=f"{settings.ziti.api.management}/edge/management/v1/authenticate?method=cert",
        status_code=200,
        json={"data": {"token": "auth-token"}},
    )
    httpx_mock.add_response(
        method="GET",
        url=f"{settings.ziti.api.management}/edge/management/v1/services/service123",
        match_headers={"zt-session": "auth-token"},
        json={"data": {"id": "service123", "name": "svc"}},
    )
    async with ZitiManagementAPI(settings) as api:
        result = await api.get("/services", "service123")
    assert result["id"] == "service123"
    assert result["name"] == "svc"
    mocked_ctx.assert_called_once_with("my-identity.json")


def test_invalid_auth(
    settings_factory: SettingsFactory,
):
    settings = settings_factory(
        ziti={
            "url": "https://ziti.example.com",
            "read_timeout": 10,
            "ssl_verify": True,
            "auth": {},
        }
    )
    with pytest.raises(ZitiAuthError) as cv:
        _ = ZitiManagementAPI(settings).auth

    assert str(cv.value) == "Unsupported authentication method for OpenZiti."


@pytest.mark.asyncio
async def test_enroll_identity(
    settings_factory: SettingsFactory,
    httpx_mock: HTTPXMock,
):
    settings = settings_factory()
    httpx_mock.add_response(
        method="POST",
        url=f"{settings.ziti.api.client}/edge/client/v1/enroll?method=ott&token=jti",
        status_code=200,
        match_content=b"csr",
        match_headers={"Content-Type": "application/x-pem-file"},
        json={"identity": "whatever"},
    )
    async with ZitiClientAPI(settings) as api:
        result = await api.enroll_identity("jti", "csr")

    assert result == {"identity": "whatever"}



import asyncio
import json
import re
from pathlib import Path
from typing import Annotated

import typer

from mrok.cli.commands.admin.utils import parse_tags
from mrok.conf import Settings
from mrok.ziti.api import ZitiClientAPI, ZitiManagementAPI
from mrok.ziti.identities import register_instance

RE_EXTENSION_ID = re.compile(r"(?i)EXT-\d{4}-\d{4}")


async def do_register(
    settings: Settings, extension_id: str, instance_id: str, tags: list[str] | None
):
    async with ZitiManagementAPI(settings) as mgmt_api, ZitiClientAPI(settings) as client_api:
        return await register_instance(
            mgmt_api, client_api, extension_id, instance_id, tags=parse_tags(tags)
        )


def validate_extension_id(extension_id: str):
    if not RE_EXTENSION_ID.fullmatch(extension_id):
        raise typer.BadParameter("ext_id must match EXT-xxxx-yyyy (case-insensitive)")
    return extension_id


def register(app: typer.Typer) -> None:
    @app.command("instance")
    def register_instance(
        ctx: typer.Context,
        extension_id: str = typer.Argument(
            ..., callback=validate_extension_id, help="Extension ID in format EXT-xxxx-yyyy"
        ),
        instance_id: str = typer.Argument(..., help="Instance ID"),
        output: Path = typer.Argument(
            ...,
            file_okay=True,
            dir_okay=False,
            writable=True,
            resolve_path=True,
            help="Output file (default: stdout)",
        ),
        tags: Annotated[
            list[str] | None,
            typer.Option(
                "--tag",
                "-t",
                help="Add tag",
                show_default=True,
            ),
        ] = None,
    ):
        """Register a new Extension Instance in OpenZiti (identity)."""
        _, identity_file = asyncio.run(do_register(ctx.obj, extension_id, instance_id, tags))
        json.dump(identity_file, output.open("w"))



import asyncio
import re

import typer

from mrok.conf import Settings
from mrok.ziti.api import ZitiManagementAPI
from mrok.ziti.services import unregister_extension

RE_EXTENSION_ID = re.compile(r"(?i)EXT-\d{4}-\d{4}")


async def do_unregister(settings: Settings, extension_id: str):
    async with ZitiManagementAPI(settings) as api:
        await unregister_extension(settings, api, extension_id)


def validate_extension_id(extension_id: str):
    if not RE_EXTENSION_ID.fullmatch(extension_id):
        raise typer.BadParameter("ext_id must match EXT-xxxx-yyyy (case-insensitive)")
    return extension_id


def register(app: typer.Typer) -> None:
    @app.command("extension")
    def unregister_extension(
        ctx: typer.Context,
        extension_id: str = typer.Argument(
            ..., callback=validate_extension_id, help="Extension ID in format EXT-xxxx-yyyy"
        ),
    ):
        """Unregister a new Extension in OpenZiti (service)."""
        asyncio.run(do_unregister(ctx.obj, extension_id))



from mrok.errors import MrokError


class ProxyIdentityNotFoundError(MrokError):
    pass


class ProxyIdentityAlreadyExistsError(MrokError):
    pass


class ServiceNotFoundError(MrokError):
    pass


class UserIdentityNotFoundError(MrokError):
    pass


class ConfigTypeNotFoundError(MrokError):
    pass


class ServiceAlreadyRegisteredError(MrokError):
    pass



from mrok.ziti.constants import MROK_SERVICE_TAG_NAME, MROK_VERSION_TAG_NAME

EXTENSION_RESPONSE = {
    "id": "5Jm3PpLQ4mdzqXNRszhE0G",
    "name": "ext-1234-5678",
    "extension": {"id": "EXT-1234-5678"},
    "tags": {"account": "ACC-5555-3333", MROK_VERSION_TAG_NAME: "1.0"},
}


INSTANCE_RESPONSE = {
    "id": "h.KUkPOyZ4",
    "name": "ins-1234-5678-0001.ext-1234-5678",
    "extension": {"id": "EXT-1234-5678"},
    "instance": {"id": "INS-1234-5678-0001"},
    "tags": {
        "account": "ACC-5555-3333",
        MROK_VERSION_TAG_NAME: "1.0",
        MROK_SERVICE_TAG_NAME: "ext-1234-5678",
    },
}

INSTANCE_CREATE_RESPONSE = {
    "id": "h.KUkPOyZ4",
    "name": "ins-1234-5678-0001.ext-1234-5678",
    "extension": {"id": "EXT-1234-5678"},
    "instance": {"id": "INS-1234-5678-0001"},
    "identity": {
        "ztAPI": "https://ziti.exts.platform.softwareone.com/edge/client/v1",
        "ztAPIs": None,
        "configTypes": None,
        "id": {
            "key": "pem:-----BEGIN PRIVATE KEY-----\n...\n-----END PRIVATE KEY-----\n",
            "cert": "pem:-----BEGIN CERTIFICATE-----\n...\n-----END CERTIFICATE-----\n",
            "ca": "pem:-----BEGIN CERTIFICATE-----\n...\n-----END CERTIFICATE-----\n",
        },
        "enableHa": None,
    },
    "tags": {
        "account": "ACC-5555-3333",
        MROK_VERSION_TAG_NAME: "1.0",
        MROK_SERVICE_TAG_NAME: "ext-1234-5678",
    },
}



import asyncio
from typing import Any

import pytest

from mrok.http.forwarder import ASGIReceive, ASGISend, ForwardAppBase


class FakeReader:
    def __init__(self, chunks: list[bytes]):
        # a single bytes buffer that we will serve from
        self._buffer = b"".join(chunks)
        self._pos = 0

    async def readline(self) -> bytes:
        # return up to and including first CRLF
        if self._pos >= len(self._buffer):
            return b""
        idx = self._buffer.find(b"\n", self._pos)
        if idx == -1:
            # return rest
            data = self._buffer[self._pos :]
            self._pos = len(self._buffer)
            return data
        idx += 1
        data = self._buffer[self._pos : idx]
        self._pos = idx
        return data

    async def read(self, n: int = -1) -> bytes:
        if self._pos >= len(self._buffer):
            return b""
        if n < 0:
            data = self._buffer[self._pos :]
            self._pos = len(self._buffer)
            return data
        data = self._buffer[self._pos : self._pos + n]
        self._pos += len(data)
        return data

    async def readexactly(self, n: int) -> bytes:
        # simplistic: if not enough, raise IncompleteReadError
        remaining = len(self._buffer) - self._pos
        if remaining < n:
            chunk = self._buffer[self._pos :]
            self._pos = len(self._buffer)
            raise asyncio.IncompleteReadError(partial=chunk, expected=n)
        data = self._buffer[self._pos : self._pos + n]
        self._pos += n
        return data


class FakeWriter:
    def __init__(self):
        self.buffer = bytearray()
        self.closed = False

    def write(self, data: bytes) -> None:
        self.buffer.extend(data)

    async def drain(self) -> None:
        await asyncio.sleep(0)  # yield control

    def close(self) -> None:
        self.closed = True

    async def wait_closed(self) -> None:
        await asyncio.sleep(0)


def send_collector(messages: list[dict[str, Any]]) -> ASGISend:
    async def _send(msg: dict[str, Any]) -> None:
        # keep async signature; no awaits required
        messages.append(msg)
        await asyncio.sleep(0)

    return _send


def make_receive(events: list[dict[str, Any]]) -> ASGIReceive:
    queue: list[dict[str, Any]] = list(events)

    async def _receive() -> dict[str, Any]:
        if not queue:
            await asyncio.sleep(0)
            return {"type": "http.request", "body": b"", "more_body": False}
        return queue.pop(0)

    return _receive


class ForwardApp(ForwardAppBase):
    """Test-local concrete ForwardApp matching the old sidecar ctor signature.

    Tests can subclass this to override select_backend when needed.
    """

    def __init__(self, target_address: str | None = None, read_chunk_size: int = 65536) -> None:
        super().__init__(read_chunk_size=read_chunk_size)
        self._target_address = target_address

    async def select_backend(self, scope, headers):
        # default dummy backend; tests will usually override this method
        return FakeReader([]), FakeWriter()


@pytest.mark.asyncio
async def test_non_http_scope_returns_500():
    app = ForwardApp("127.0.0.1:8000")

    sent = []
    send = send_collector(sent)

    scope = {"type": "websocket"}
    receive = make_receive([])

    await app(scope, receive, send)

    assert any(m.get("type") == "http.response.start" and m.get("status") == 500 for m in sent)
    assert any(b"Unsupported" in m.get("body", b"") for m in sent)


@pytest.mark.asyncio
async def test_empty_status_line_returns_502():
    class App(ForwardApp):
        async def select_backend(self, scope, headers):
            # reader that immediately returns empty status line
            return FakeReader([b""]), FakeWriter()

    app = App("127.0.0.1:8000")
    sent = []
    send = send_collector(sent)
    scope = {"type": "http"}
    receive = make_receive([])

    await app(scope, receive, send)

    assert any(m.get("status") == 502 for m in sent)


@pytest.mark.asyncio
async def test_chunked_request_and_chunked_response():
    # backend will echo chunked response "hello"
    status = b"HTTP/1.1 200 OK\r\n"
    headers = b"transfer-encoding: chunked\r\n\r\n"
    body = b"5\r\nhello\r\n0\r\n\r\n"
    reader = FakeReader([status, headers, body])
    writer = FakeWriter()

    class App(ForwardApp):
        async def select_backend(self, scope, headers):
            return reader, writer

    app = App("127.0.0.1:8000")

    # ASGI receive will provide two body chunks
    events = [
        {"type": "http.request", "body": b"ab", "more_body": True},
        {"type": "http.request", "body": b"cd", "more_body": False},
    ]
    receive = make_receive(events)
    sent = []
    send = send_collector(sent)
    scope = {
        "type": "http",
        "method": "POST",
        "path": "/up",
        "query_string": b"",
        "headers": [],
    }

    await app(scope, receive, send)

    # verify writer got chunked request framing: two chunks and final 0
    buf = bytes(writer.buffer)
    assert b"POST /up HTTP/1.1\r\n" in buf
    assert b"transfer-encoding: chunked\r\n" in buf
    assert b"2\r\nabcd\r\n" in buf or b"2\r\nab\r\n2\r\ncd\r\n" in buf
    assert b"0\r\n\r\n" in buf

    # verify sends: start + body chunk + final
    types = [m["type"] for m in sent]
    assert "http.response.start" in types
    bodies = [m for m in sent if m["type"] == "http.response.body"]
    assert bodies[0]["body"] == b"hello"
    assert bodies[-1]["more_body"] is False


@pytest.mark.asyncio
async def test_write_skips_expect_header():
    app = ForwardApp("127.0.0.1:8000")
    writer = FakeWriter()
    headers = [(b"expect", b"100-continue"), (b"host", b"example")]
    await app.write_request_line_and_headers(writer, "GET", "/", headers)
    buf = bytes(writer.buffer)
    assert b"expect" not in buf.lower()


@pytest.mark.asyncio
async def test_stream_request_handles_disconnect():
    app = ForwardApp("127.0.0.1:8000")
    writer = FakeWriter()
    receive = make_receive([{"type": "http.disconnect"}])
    await app.stream_request_body(receive, writer, use_chunked=True)
    assert writer.closed is True


@pytest.mark.asyncio
async def test_read_status_non_digit_returns_502():
    app = ForwardApp("127.0.0.1:8000")
    reader = FakeReader([b"HTTP/1.1 BADSTATUS\r\n", b"\r\n"])
    status, headers, raw = await app.read_status_and_headers(reader, b"HTTP/1.1 BADSTATUS\r\n")
    assert status == 502


@pytest.mark.asyncio
async def test_chunked_invalid_size_line_sends_final_only():
    app = ForwardApp("127.0.0.1:8000")
    reader = FakeReader([b"not-hex\r\n"])
    sent = []
    send = send_collector(sent)
    await app.stream_response_body(reader, send, {b"transfer-encoding": b"chunked"})
    # Only the terminating empty body should be sent
    bodies = [m for m in sent if m["type"] == "http.response.body"]
    assert len(bodies) == 1
    assert bodies[0]["more_body"] is False


@pytest.mark.asyncio
async def test_content_length_invalid_parsed_as_stream_until_eof():
    app = ForwardApp("127.0.0.1:8000", read_chunk_size=4)
    reader = FakeReader([b"abcd"])
    sent = []
    send = send_collector(sent)
    await app.stream_response_body(reader, send, {b"content-length": b"notint"})
    bodies = [m for m in sent if m["type"] == "http.response.body"]
    # should forward the bytes and then final empty
    assert any(b["body"] == b"abcd" for b in bodies)
    assert bodies[-1]["more_body"] is False


@pytest.mark.asyncio
async def test_stream_request_handles_disconnect_non_chunked():
    # Verify non-chunked path also closes writer on disconnect
    app = ForwardApp("127.0.0.1:8000")
    writer = FakeWriter()
    receive = make_receive([{"type": "http.disconnect"}])
    await app.stream_request_body(receive, writer, use_chunked=False)
    assert writer.closed is True


@pytest.mark.asyncio
async def test_chunked_response_eof_immediately_sends_final():
    # If reader.readline() returns empty immediately, chunked path should send only final empty body
    app = ForwardApp("127.0.0.1:8000")
    reader = FakeReader([])  # empty buffer -> first readline() returns b""
    sent = []
    send = send_collector(sent)
    await app.stream_response_body(reader, send, {b"transfer-encoding": b"chunked"})
    bodies = [m for m in sent if m["type"] == "http.response.body"]
    assert len(bodies) == 1
    assert bodies[0]["more_body"] is False


@pytest.mark.asyncio
async def test_content_length_partial_eof_results_in_final_empty():
    # content-length header present but reader provides fewer bytes than expected
    # -> final empty send
    app = ForwardApp("127.0.0.1:8000", read_chunk_size=4)
    # simulate reader that will return 2 bytes then EOF
    reader = FakeReader([b"ab"])  # content-length says 5, but only 2 provided
    sent = []
    send = send_collector(sent)
    await app.stream_response_body(reader, send, {b"content-length": b"5"})
    bodies = [m for m in sent if m["type"] == "http.response.body"]
    # we should have forwarded the 'ab' and then sent a final empty body
    assert any(b["body"] == b"ab" for b in bodies)
    assert bodies[-1]["more_body"] is False


@pytest.mark.asyncio
async def test_content_length_forwarding_and_response():
    # backend will return content-length response
    status = b"HTTP/1.1 200 OK\r\n"
    headers = b"content-length: 11\r\n\r\n"
    body = b"hello world"
    reader = FakeReader([status, headers, body])
    writer = FakeWriter()

    class App(ForwardApp):
        async def select_backend(self, scope, headers):
            return reader, writer

    # include content-length header in incoming headers so we avoid chunked request
    app = App("127.0.0.1:8000")
    events = [
        {"type": "http.request", "body": b"hello world", "more_body": False},
    ]
    receive = make_receive(events)
    sent = []
    send = send_collector(sent)
    scope = {
        "type": "http",
        "method": "POST",
        "path": "/cl",
        "query_string": b"",
        "headers": [(b"content-length", b"11")],
    }

    await app(scope, receive, send)

    buf = bytes(writer.buffer)
    # should not contain chunk framing
    assert b"transfer-encoding" not in buf
    assert b"hello world" in buf

    bodies = [m for m in sent if m["type"] == "http.response.body"]
    assert bodies[0]["body"] == b"hello world"
    assert bodies[-1]["more_body"] is False


@pytest.mark.asyncio
async def test_stream_request_body_chunked_empty_final():
    # When chunked and the only event has empty body and more_body False,
    # the writer should receive only the terminating 0 chunk.
    app = ForwardApp("127.0.0.1:8000")
    writer = FakeWriter()
    receive = make_receive([{"type": "http.request", "body": b"", "more_body": False}])
    await app.stream_request_body(receive, writer, use_chunked=True)
    buf = bytes(writer.buffer)
    assert b"0\r\n\r\n" in buf


@pytest.mark.asyncio
async def test_stream_request_body_non_chunked_writes_body():
    # Non-chunked bodies should be written raw to the writer
    app = ForwardApp("127.0.0.1:8000")
    writer = FakeWriter()
    receive = make_receive([{"type": "http.request", "body": b"xyz", "more_body": False}])
    await app.stream_request_body(receive, writer, use_chunked=False)
    assert bytes(writer.buffer) == b"xyz"


@pytest.mark.asyncio
async def test_stream_request_body_chunked_with_unknown_event():
    # If an unknown event type is received, the loop should continue until http.request
    app = ForwardApp("127.0.0.1:8000")
    writer = FakeWriter()
    events = [
        {"type": "something-else"},
        {"type": "http.request", "body": b"x", "more_body": False},
    ]
    receive = make_receive(events)
    await app.stream_request_body(receive, writer, use_chunked=True)
    buf = bytes(writer.buffer)
    assert b"1\r\nx\r\n" in buf


@pytest.mark.asyncio
async def test_stream_request_body_non_chunked_multiple_events():
    # Non-chunked path with multiple http.request events where first has more_body True
    app = ForwardApp("127.0.0.1:8000")
    writer = FakeWriter()
    events = [
        {"type": "http.request", "body": b"a", "more_body": True},
        {"type": "http.request", "body": b"b", "more_body": False},
    ]
    receive = make_receive(events)
    await app.stream_request_body(receive, writer, use_chunked=False)
    assert bytes(writer.buffer) == b"ab"


@pytest.mark.asyncio
async def test_stream_request_body_non_chunked_empty_final():
    # Non-chunked with single empty body and more_body False should not write and should break
    app = ForwardApp("127.0.0.1:8000")
    writer = FakeWriter()
    receive = make_receive([{"type": "http.request", "body": b"", "more_body": False}])
    await app.stream_request_body(receive, writer, use_chunked=False)
    assert bytes(writer.buffer) == b""


@pytest.mark.asyncio
async def test_stream_request_body_non_chunked_with_unknown_event():
    # Unknown events should be ignored and the loop should continue until http.request
    app = ForwardApp("127.0.0.1:8000")
    writer = FakeWriter()
    events = [
        {"type": "ignored-event"},
        {"type": "http.request", "body": b"z", "more_body": False},
    ]
    receive = make_receive(events)
    await app.stream_request_body(receive, writer, use_chunked=False)
    assert bytes(writer.buffer) == b"z"


@pytest.mark.asyncio
async def test_stream_until_eof_response():
    # backend returns no content-length and no transfer-encoding, so we stream until EOF
    status = b"HTTP/1.1 200 OK\r\n"
    headers = b"\r\n"
    body = b"streamed-data"
    reader = FakeReader([status, headers, body])
    writer = FakeWriter()

    class App(ForwardApp):
        async def select_backend(self, scope, headers):
            return reader, writer

    app = App("127.0.0.1:8000")
    receive = make_receive([])
    sent = []
    send = send_collector(sent)
    scope = {
        "type": "http",
        "method": "GET",
        "path": "/stream",
        "query_string": b"",
        "headers": [],
    }

    await app(scope, receive, send)

    bodies = [m for m in sent if m["type"] == "http.response.body"]
    assert any(b["body"] == b"streamed-data" for b in bodies)
    assert bodies[-1]["more_body"] is False


def test_format_and_host_helpers():
    app = ForwardApp("127.0.0.1:8000")
    # raw_path
    scope1 = {"raw_path": b"/raw/path"}
    assert app.format_path(scope1) == "/raw/path"

    # path + query
    scope2 = {"path": "/p", "query_string": b"a=1"}
    assert app.format_path(scope2) == "/p?a=1"

    # ensure_host_header does nothing when host present
    headers = [(b"host", b"example")]
    out = app.ensure_host_header(list(headers), {})
    assert any(h[0] == b"host" for h in out)

    # ensure_host_header adds host when server present
    out2 = app.ensure_host_header([], {"server": "127.0.0.1:8000"})
    assert any(h[0] == b"host" for h in out2)


@pytest.mark.asyncio
async def test_incoming_transfer_encoding_preserved():
    # If the incoming ASGI headers already include Transfer-Encoding, we must not add another
    status = b"HTTP/1.1 200 OK\r\n"
    headers = b"transfer-encoding: chunked\r\n\r\n"
    body = b"1\r\na\r\n0\r\n\r\n"
    reader = FakeReader([status, headers, body])
    writer = FakeWriter()

    class App(ForwardApp):
        async def select_backend(self, scope, headers):
            return reader, writer

    app = App("127.0.0.1:8000")

    events = [
        {"type": "http.request", "body": b"a", "more_body": False},
    ]
    receive = make_receive(events)
    sent = []
    send = send_collector(sent)
    scope = {
        "type": "http",
        "method": "POST",
        "path": "/up",
        "query_string": b"",
        "headers": [(b"transfer-encoding", b"chunked")],
    }

    await app(scope, receive, send)

    buf = bytes(writer.buffer)
    # transfer-encoding must be present but not duplicated
    assert buf.count(b"transfer-encoding: chunked") >= 1


@pytest.mark.asyncio
async def test_chunked_response_with_trailer():
    # backend sends a chunked response that includes trailer headers
    status = b"HTTP/1.1 200 OK\r\n"
    headers = b"transfer-encoding: chunked\r\n\r\n"
    # chunked body "hello" with a trailer header 'x-trailer: v'
    body = b"5\r\nhello\r\n0\r\nx-trailer: v\r\n\r\n"
    reader = FakeReader([status, headers, body])
    writer = FakeWriter()

    class App(ForwardApp):
        async def select_backend(self, scope, headers):
            return reader, writer

    app = App("127.0.0.1:8000")
    receive = make_receive([])
    sent = []
    send = send_collector(sent)
    scope = {
        "type": "http",
        "method": "GET",
        "path": "/tr",
        "query_string": b"",
        "headers": [],
    }

    await app(scope, receive, send)

    bodies = [m for m in sent if m["type"] == "http.response.body"]
    assert any(b["body"] == b"hello" for b in bodies)
    assert bodies[-1]["more_body"] is False


@pytest.mark.asyncio
async def test_malformed_header_line_parsing():
    # backend returns a malformed header line (no colon) plus a valid content-length
    status = b"HTTP/1.1 200 OK\r\n"
    headers = b"bad-header-line\r\ncontent-length: 3\r\n\r\n"
    body = b"xyz"
    reader = FakeReader([status, headers, body])
    writer = FakeWriter()

    class App(ForwardApp):
        async def select_backend(self, scope, headers):
            return reader, writer

    app = App("127.0.0.1:8000")
    receive = make_receive([])
    sent = []
    send = send_collector(sent)
    scope = {
        "type": "http",
        "method": "GET",
        "path": "/bad",
        "query_string": b"",
        "headers": [],
    }

    await app(scope, receive, send)

    bodies = [m for m in sent if m["type"] == "http.response.body"]
    # the body should still be forwarded despite the malformed header line
    assert any(b["body"] == b"xyz" for b in bodies)
    assert bodies[-1]["more_body"] is False



import pytest
from pytest_mock import MockerFixture

from mrok.agent.sidecar import main


@pytest.mark.parametrize("target_addr", ["/path/to/unix.sock", ("localhost", 8080)])
def test_run_sidecar(mocker: MockerFixture, target_addr: str | tuple[str, int]):
    mocked_app = mocker.MagicMock()
    mocked_app_ctor = mocker.patch(
        "mrok.agent.sidecar.main.ForwardApp",
        return_value=mocked_app,
    )

    mocked_config = mocker.MagicMock()
    mocked_config_ctor = mocker.patch(
        "mrok.agent.sidecar.main.MrokBackendConfig",
        return_value=mocked_config,
    )

    mocked_server = mocker.MagicMock()
    mocked_server_ctor = mocker.patch(
        "mrok.agent.sidecar.main.MrokServer",
        return_value=mocked_server,
    )

    main.run_sidecar("ziti-identity.json", target_addr)

    mocked_config_ctor.assert_called_once_with(mocked_app, "ziti-identity.json")
    mocked_app_ctor.assert_called_once_with(target_addr)
    mocked_server_ctor.assert_called_once_with(mocked_config)
    mocked_server.run.assert_called_once()


def test_run(mocker: MockerFixture):
    mocked_start_fn = mocker.MagicMock()
    mocked_partial = mocker.patch("mrok.agent.sidecar.main.partial", return_value=mocked_start_fn)

    mocked_master = mocker.MagicMock()
    mocked_master_ctor = mocker.patch(
        "mrok.agent.sidecar.main.Master",
        return_value=mocked_master,
    )

    main.run("ziti-identity.json", "target-addr", workers=10, reload=True)

    mocked_partial.assert_called_once_with(
        main.run_sidecar,
        "ziti-identity.json",
        "target-addr",
    )
    mocked_master_ctor.assert_called_once_with(mocked_start_fn, workers=10, reload=True)
    mocked_master.run.assert_called_once()



from rich.console import Console
from rich.highlighter import ReprHighlighter
from rich.theme import Theme


class MrokHighlighter(ReprHighlighter):
    prefixes = ("EXT", "ext", "INS", "ins")
    highlights = ReprHighlighter.highlights + [
        rf"(?P<mrok_id>(?:{'|'.join(prefixes)})(?:-\d{{4}})*)"
    ]


def get_console(stderr: bool = False) -> Console:
    return Console(
        stderr=stderr,
        highlighter=MrokHighlighter(),
        theme=Theme({"repr.mrok_id": "bold light_salmon3"}),
    )



import multiprocessing
from collections.abc import Callable
from typing import Annotated, Any

import typer
from gunicorn.app.base import BaseApplication

from mrok.controller.app import app as asgi_app
from mrok.logging import get_logging_config


def number_of_workers():
    return (multiprocessing.cpu_count() * 2) + 1


class StandaloneApplication(BaseApplication):  # pragma: no cover
    def __init__(self, application: Callable, options: dict[str, Any] | None = None):
        self.options = options or {}
        self.application = application
        super().__init__()

    def load_config(self):
        config = {
            key: value
            for key, value in self.options.items()
            if key in self.cfg.settings and value is not None
        }
        for key, value in config.items():
            self.cfg.set(key.lower(), value)

    def load(self):
        return self.application


default_workers = number_of_workers()


def register(app: typer.Typer) -> None:
    @app.command("run")
    def run_controller(
        ctx: typer.Context,
        host: Annotated[
            str,
            typer.Option(
                "--host",
                "-h",
                help="Host to bind to. Default: 127.0.0.1",
                show_default=True,
            ),
        ] = "127.0.0.1",
        port: Annotated[
            int,
            typer.Option(
                "--port",
                "-p",
                help="Port to bind to. Default: 8000",
                show_default=True,
            ),
        ] = 8000,
        workers: Annotated[
            int,
            typer.Option(
                "--workers",
                "-w",
                help=f"Number of workers. Default: {default_workers}",
                show_default=True,
            ),
        ] = default_workers,
        dev: Annotated[
            bool,
            typer.Option(
                "--reload",
                "-r",
                help="Enable auto-reload. Default: False",
                show_default=True,
            ),
        ] = False,
    ):
        """Run the mrok controller with Gunicorn and Uvicorn workers."""
        options = {
            "bind": f"{host}:{port}",
            "workers": workers,
            "worker_class": "uvicorn_worker.UvicornWorker",
            "logconfig_dict": get_logging_config(ctx.obj),
            "reload": dev,
        }
        StandaloneApplication(asgi_app, options).run()



import pytest
from pytest_mock import MockerFixture
from typer.testing import CliRunner

from mrok.cli import app, main, run


def test_help_show_banner(mocker: MockerFixture):
    mocked_show = mocker.patch("mrok.cli.main.show_banner")
    runner = CliRunner()
    # Run the command
    runner.invoke(
        app,
        ["--help"],
    )
    mocked_show.assert_called_once()


def test_run(mocker: MockerFixture):
    app = mocker.MagicMock()
    mocker.patch.object(main, "app", app)
    run()
    app.assert_called_once()


def test_run_exception(mocker: MockerFixture):
    app = mocker.MagicMock()
    err_console = mocker.MagicMock()
    app.side_effect = Exception("whatever")
    mocker.patch.object(main, "app", app)
    mocker.patch.object(main, "err_console", err_console)
    with pytest.raises(SystemExit) as exc:
        run()
    assert exc.value.code == -1
    err_console.print.assert_called_once_with("[bold red]Error:[/bold red] whatever")



import logging
from functools import partial

import fastapi_pagination
from fastapi import Depends, FastAPI
from fastapi.routing import APIRoute, APIRouter

from mrok.conf import get_settings
from mrok.controller.auth import authenticate
from mrok.controller.openapi import generate_openapi_spec
from mrok.controller.routes import router as extensions_router

logger = logging.getLogger(__name__)


tags_metadata = [
    {
        "name": "Extensions",
        "description": "Manage Extensions (services).",
    },
    {
        "name": "Instances",
        "description": "Manage Extension Instances (identities).",
    },
]


def setup_custom_serialization(router: APIRouter):
    for api_route in router.routes:
        if (
            isinstance(api_route, APIRoute)
            and hasattr(api_route, "response_model")
            and api_route.response_model
        ):
            api_route.response_model_exclude_none = True


def setup_app():
    app = FastAPI(
        title="mrok Controller API",
        description="API to orchestrate OpenZiti for Extensions.",
        swagger_ui_parameters={"showExtensions": False, "showCommonExtensions": False},
        openapi_tags=tags_metadata,
        version="5.0.0",
        root_path="/public/v1",
    )
    fastapi_pagination.add_pagination(app)

    setup_custom_serialization(extensions_router)

    # TODO: Add healthcheck
    app.include_router(
        extensions_router, prefix="/extensions", dependencies=[Depends(authenticate)]
    )

    settings = get_settings()

    app.openapi = partial(generate_openapi_spec, app, settings)
    return app


app = setup_app()



import logging

from uvicorn.config import Config
from uvicorn.lifespan.on import LifespanOn


class MrokLifespan(LifespanOn):
    def __init__(self, config: Config) -> None:
        super().__init__(config)
        self.logger = logging.getLogger("mrok.proxy")



import asyncio
import json
import logging
from pathlib import Path
from typing import Annotated, Any

import typer

from mrok.cli.commands.admin.utils import parse_tags
from mrok.conf import Settings
from mrok.ziti.api import TagsType, ZitiClientAPI, ZitiManagementAPI
from mrok.ziti.bootstrap import bootstrap_identity

logger = logging.getLogger(__name__)


async def bootstrap(
    settings: Settings, forced: bool, tags: TagsType | None
) -> tuple[str, dict[str, Any] | None]:
    async with ZitiManagementAPI(settings) as mgmt_api, ZitiClientAPI(settings) as client_api:
        return await bootstrap_identity(
            mgmt_api,
            client_api,
            settings.proxy.identity,
            settings.proxy.mode,
            forced,
            tags,
        )


def register(app: typer.Typer) -> None:
    @app.command("bootstrap")
    def run_bootstrap(
        ctx: typer.Context,
        identity_file: Path = typer.Argument(
            Path("proxy_identity.json"),
            help="Path to identity output file",
            writable=True,
        ),
        forced: bool = typer.Option(
            False,
            "--force",
            help="Regenerate identity even if it already exists",
        ),
        tags: Annotated[
            list[str] | None,
            typer.Option(
                "--tag",
                "-t",
                help="Add tag",
                show_default=True,
            ),
        ] = None,
    ):
        """Run the mrok bootstrap."""
        _, identity_json = asyncio.run(bootstrap(ctx.obj, forced, parse_tags(tags)))
        if identity_json:
            json.dump(identity_json, identity_file.open("w"))



import signal
from collections.abc import Generator
from pathlib import Path

from pytest_mock import MockerFixture
from watchfiles import Change

from mrok.http.master import Master


def test_init(mocker: MockerFixture):
    mocked_setup_signals = mocker.patch.object(Master, "setup_signals_handler")
    mocked_start_fn = mocker.MagicMock()
    master = Master(mocked_start_fn, 3, False)
    assert master.start_fn == mocked_start_fn
    assert master.workers == 3
    assert master.reload is False
    mocked_setup_signals.assert_called_once()


def test_setup_signals_handler(mocker: MockerFixture):
    mocked_signal = mocker.patch("mrok.http.master.signal.signal")
    master = Master(mocker.MagicMock(), 3, False)
    assert mocked_signal.call_count == 2
    assert mocked_signal.mock_calls[0].args == (signal.SIGINT, master.handle_signal)
    assert mocked_signal.mock_calls[1].args == (signal.SIGTERM, master.handle_signal)


def test_handle_signal(mocker: MockerFixture):
    master = Master(mocker.MagicMock(), 3, False)
    assert master.stop_event.is_set() is False
    master.handle_signal()
    assert master.stop_event.is_set() is True


def test_start(mocker: MockerFixture):
    processes = [mocker.MagicMock(), mocker.MagicMock(), mocker.MagicMock()]
    mocked_start_process = mocker.patch(
        "mrok.http.master.start_process",
        side_effect=processes,
    )
    start_fn = mocker.MagicMock()
    master = Master(start_fn, 3, False)
    master.start()
    assert master.worker_processes == processes
    for i in range(3):
        assert mocked_start_process.mock_calls[i].args == (start_fn, "function", (), None)


def test_stop(mocker: MockerFixture):
    master = Master(mocker.MagicMock(), 3, False)
    p1 = mocker.MagicMock()
    p2 = mocker.MagicMock()
    master.worker_processes = [p1, p2]
    master.stop()
    p1.stop.assert_called_once_with(sigint_timeout=5, sigkill_timeout=1)
    p2.stop.assert_called_once_with(sigint_timeout=5, sigkill_timeout=1)


def test_restart(mocker: MockerFixture):
    mocked_stop = mocker.patch.object(Master, "stop")
    mocked_start = mocker.patch.object(Master, "start")
    master = Master(mocker.MagicMock(), 3, False)
    master.restart()
    mocked_stop.assert_called()
    mocked_start.assert_called()


def test_iter(mocker: MockerFixture):
    master = Master(mocker.MagicMock(), 3, False)
    assert iter(master) == master


def test_next(mocker: MockerFixture):
    def watcher() -> Generator:
        yield {(Change.modified, "/file1.py")}
        yield None

    master = Master(mocker.MagicMock(), 3, False)
    master.watcher = watcher()
    assert next(master) == [Path("/file1.py")]
    assert next(master) is None


def test_run(mocker: MockerFixture):
    mocked_start = mocker.patch.object(Master, "start")
    master = Master(mocker.MagicMock(), 3, False)
    mocked_stop_event = mocker.MagicMock()
    master.stop_event = mocked_stop_event
    master.run()
    mocked_start.assert_called_once()
    mocked_stop_event.wait.assert_called_once()


def test_run_with_reload(mocker: MockerFixture):
    mocker.patch.object(Master, "start")
    mocked_restart = mocker.patch.object(Master, "restart")

    def watcher() -> Generator:
        yield {(Change.modified, "/file1.py")}
        yield None

    master = Master(mocker.MagicMock(), 3, True)
    master.watcher = watcher()
    master.run()

    mocked_restart.assert_called_once()



import asyncio
from typing import Any

import pytest
from pytest_mock import MockerFixture

from mrok.agent.sidecar.app import ASGIReceive, ASGISend, ForwardApp


class FakeReader:
    def __init__(self, chunks: list[bytes]):
        # a single bytes buffer that we will serve from
        self._buffer = b"".join(chunks)
        self._pos = 0

    async def readline(self) -> bytes:
        # return up to and including first CRLF
        if self._pos >= len(self._buffer):
            return b""
        idx = self._buffer.find(b"\n", self._pos)
        if idx == -1:
            # return rest
            data = self._buffer[self._pos :]
            self._pos = len(self._buffer)
            return data
        idx += 1
        data = self._buffer[self._pos : idx]
        self._pos = idx
        return data

    async def read(self, n: int = -1) -> bytes:
        if self._pos >= len(self._buffer):
            return b""
        if n < 0:
            data = self._buffer[self._pos :]
            self._pos = len(self._buffer)
            return data
        data = self._buffer[self._pos : self._pos + n]
        self._pos += len(data)
        return data

    async def readexactly(self, n: int) -> bytes:
        # simplistic: if not enough, raise IncompleteReadError
        remaining = len(self._buffer) - self._pos
        if remaining < n:
            chunk = self._buffer[self._pos :]
            self._pos = len(self._buffer)
            raise asyncio.IncompleteReadError(partial=chunk, expected=n)
        data = self._buffer[self._pos : self._pos + n]
        self._pos += n
        return data


class FakeWriter:
    def __init__(self):
        self.buffer = bytearray()
        self.closed = False

    def write(self, data: bytes) -> None:
        self.buffer.extend(data)

    async def drain(self) -> None:
        await asyncio.sleep(0)  # yield control

    def close(self) -> None:
        self.closed = True

    async def wait_closed(self) -> None:
        await asyncio.sleep(0)


def send_collector(messages: list[dict[str, Any]]) -> ASGISend:
    async def _send(msg: dict[str, Any]) -> None:
        # keep async signature; no awaits required
        messages.append(msg)
        await asyncio.sleep(0)

    return _send


def make_receive(events: list[dict[str, Any]]) -> ASGIReceive:
    queue: list[dict[str, Any]] = list(events)

    async def _receive() -> dict[str, Any]:
        if not queue:
            await asyncio.sleep(0)
            return {"type": "http.request", "body": b"", "more_body": False}
        return queue.pop(0)

    return _receive


@pytest.mark.asyncio
async def test_select_backend_paths(mocker: MockerFixture):
    # Patch asyncio open_connection and open_unix_connection to return FakeReader/Writer
    async def fake_open_connection(host, port):
        await asyncio.sleep(0)
        return FakeReader([b"HTTP/1.1 200 OK\r\n\r\n", b"ok"]), FakeWriter()

    async def fake_open_unix_connection(path):
        await asyncio.sleep(0)
        return FakeReader([b"HTTP/1.1 200 OK\r\n\r\n", b"ok"]), FakeWriter()

    mocker.patch("asyncio.open_connection", new=fake_open_connection)
    mocker.patch("asyncio.open_unix_connection", new=fake_open_unix_connection)

    app_tcp = ForwardApp(("127.0.0.1", 9000))
    r1, w1 = await app_tcp.select_backend({}, {})
    assert isinstance(r1, FakeReader)

    app_unix = ForwardApp("/tmp/sock")
    r2, w2 = await app_unix.select_backend({}, {})
    assert isinstance(r2, FakeReader)



import typer

from mrok.cli.commands.admin.unregister import extensions, instances

app = typer.Typer(name="unregister", help="Unregister resources from OpenZiti.")

extensions.register(app)
instances.register(app)
